# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import sys
import traceback
import os
import copy
# choosing multiprocessing over threading for clean Control-C interrupts (provides terminate())
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from multiprocessing import Process, Manager
from vcstools.vcs_base import VcsError


class MultiProjectException(Exception):
    pass


def samefile(f1, f2):
    """
    Test whether two pathnames reference the same actual file
    This is a workaround for the fact that some platforms
    do not have os.path.samefile (particularly windows). This
    is the patch that was integrated in python 3.0 (at which
    time we can probably remove this workaround).
    """
    try:
        return os.path.samefile(f1,f2)
    except AttributeError:
        try:
            from nt import _getfinalpathname
            return _getfinalpathname(f1) == _getfinalpathname(f2)
        except (NotImplementedError, ImportError):
            # On Windows XP and earlier, two files are the same if their
            #  absolute pathnames are the same.
            # Also, on other operating systems, fake this method with a
            #  Windows-XP approximation.
            return os.path.abspath(f1) == os.path.abspath(f2)


def conditional_abspath(uri):
    """
    @param uri: The uri to check
    @return: abspath(uri) if local path otherwise pass through uri
    """
    uri2 = urlparse(uri)
    # maybe it's a local file?
    if uri2.scheme == '':
        return os.path.abspath(uri)
    else:
        return uri


def is_web_uri(source_uri):
    """
    Uses heuristics to check whether uri is a web uri (as opposed to a file path)
    :param source_uri: string representing web uri or file path
    :returns: bool
    """
    if source_uri is None or source_uri == '':
        return False
    parsed_uri = urlparse(source_uri)
    if (parsed_uri.scheme == '' and
        parsed_uri.netloc == '' and
        not '@' in parsed_uri.path.split('/')[0]):

        return False
    return True


def normalize_uri(source_uri, base_path):
    """
    If source_uri is none or a web uir, return it.
    If source_uri is a relative path, make it an absolute path.
    Else return it normalized
    :param source_uri: some uri to a file, folder, or web resource
    :param base_path: path to use to make relative paths absolute
    :returns: normalized string
    """
    if source_uri is not None and not is_web_uri(source_uri):
        if os.path.isabs(source_uri):
            source_uri = os.path.normpath(source_uri)
        else:
            source_uri2 = os.path.normpath(os.path.join(base_path, source_uri))
            print("Warning: Converted relative uri path %s to abspath %s"%(source_uri, source_uri2))
            source_uri = source_uri2
    return source_uri


def string_diff(str1_orig, str2_orig, maxlen=11, backtrack=7):
    """
    Compares strings, returns a part of str2 depending on how many
    chars into the string the first difference can be found. If the
    difference is after maxlen, a prefix of str is removed so that
    only the 'backtrack'-last letters of the common prefix remain in str2.

    This only makes sense if str1 != str2, really.

    The purpose is to print str1 -> str2 without repeating a same long prefix

    :returns: a representation of where str2 differs from str1.
    """
    result = str2_orig or ''
    if str1_orig is not None and str2_orig is not None:
        # we cannot be sure we have strings, might be lists,
        # gracefully fail convert to string
        str1 = str(str1_orig)
        str2 = str(str2_orig)
        result = str2

        if len(str2) > len(str1):
            str1 = str1.ljust(len(str2))
        charcompare = [x[0] == x[1] for x in zip(str(str2), str(str1))]
        if False in charcompare:
            commonprefix = str2[:charcompare.index(False)]
            if len(commonprefix) > maxlen:
                result = "...%s"%str2[len(commonprefix) - backtrack:]
    return result

def normabspath(localname, path):
    """
    if localname is absolute, return it normalized. If relative,
    return normalized join of path and localname
    """
    if os.path.isabs(localname) or path is None:
        return os.path.realpath(localname)
    abs_path = os.path.realpath(os.path.join(path, localname))
    return abs_path


def realpath_relation(abspath1, abspath2):
    """
    Computes the relationship abspath1 to abspath2
    :returns: None, 'SAME_AS', 'PARENT_OF', 'CHILD_OF'
    """
    assert os.path.isabs(abspath1), "Bug, %s is not absolute path"%abspath1
    assert os.path.isabs(abspath2), "Bug, %s is not absolute path"%abspath2
    realpath1 = os.path.realpath(abspath1)
    realpath2 = os.path.realpath(abspath2)
    if os.path.dirname(realpath1) == os.path.dirname(realpath2):
        if os.path.basename(realpath1) == os.path.basename(realpath2):
            return 'SAME_AS'
        return None
    else:
        commonprefix = os.path.commonprefix([realpath1, realpath2])
        if commonprefix == realpath1:
            return 'PARENT_OF'
        elif commonprefix == realpath2:
            return 'CHILD_OF'
    return None


def select_element(elements, localname):
    """
    selects entry among elements where path or localname matches.
    Prefers localname matches in case of ambiguity.
    """
    path_candidate = None
    if localname is not None:
        realpath = os.path.realpath(localname)
        for element in elements:
            if localname == element.get_local_name():
                path_candidate = element
                break
            elif realpath == os.path.realpath(element.get_path()):
                path_candidate = element
    return path_candidate


def select_elements(config, localnames):
    """
    selects config elements with given localnames, returns in the
    order given in config If localnames has one element which is path
    of the config, return all elements
    """
    if config is None:
        return []
    if localnames is None:
        return config.get_config_elements()
    elements = config.get_config_elements()
    selected = []
    notfound = []
    for localname in localnames:
        element = select_element(elements, localname)
        if element is not None:
            selected.append(element)
        else:
            notfound.append(localname)
    if notfound != []:
        # if we just passed workspace path, return all workspace entries
        if (len(localnames) == 1 and
            os.path.realpath(localnames[0]) == os.path.realpath(config.get_base_path())):

            return config.get_config_elements()
        raise MultiProjectException("Unknown elements '%s'"%notfound)
    result = []
    # select in order and remove duplicates
    for element in config.get_config_elements():
        if element in selected:
            result.append(element)
    return result


## Multithreading The following classes help with distributing work
## over several instances, providing wrapping for starting, joining,
## collecting results, and catching Exceptions. Also they provide
## support for running groups of threads sequentially, for the case
## that some library is not thread-safe.


class WorkerThread(Process):

    def __init__(self, worker, outlist, index):
        Process.__init__(self)
        self.worker = worker
        if worker is None or worker.element is None:
            raise MultiProjectException("Bug: Invalid Worker")
        self.outlist = outlist
        self.index = index

    def run(self):
        result = {}
        try:
            result = {'entry': self.worker.element.get_path_spec()}
            result_dict = self.worker.do_work()
            if result_dict is not None:
                result.update(result_dict)
            else:
                result.update({'error': MultiProjectException("worker returned None")})
        except MultiProjectException as mpe:
            result.update({'error': mpe})
        except VcsError as vcse:
            result.update({'error': vcse})
        except OSError as ose:
            result.update({'error': ose})
        except Exception as exc:
            # this would be a bug, and we need trace to find them in
            # multithreaded cases.
            traceback.print_exc(file=sys.stderr)
            result.update({'error': exc})
        self.outlist[self.index] = result


class DistributedWork():

    def __init__(self, capacity, num_threads=10, silent=True):
         # need managed array since we need the results later
        man = Manager()
        self.outputs = man.list([None for _ in range(capacity)])
        self.threads = []
        self.sequentializers = {}
        self.index = 0
        self.num_threads = num_threads
        self.silent = silent

    def add_thread(self, worker):
        thread = WorkerThread(worker, self.outputs, self.index)
        if self.index >= len(self.outputs):
            raise MultiProjectException(
                "Bug: Declared capacity exceeded %s >= %s"%(self.index,
                                                            len(self.outputs)))
        self.index += 1
        self.threads.append(thread)

    def run(self):
        """
        Execute all collected workers, terminate all on KeyboardInterrupt
        """
        if self.threads == []:
            return []
        if (self.num_threads == 1):
            for thread in self.threads:
                thread.run()
        else:
            # The following code is rather delicate and may behave differently
            # using threading or multiprocessing. running_threads is
            # intentionally not used as a shrinking list because of al the
            # possible multithreading / interruption corner cases
            # Not using Pool because of KeyboardInterrupt cases
            try:
                waiting_index = 0
                maxthreads = self.num_threads
                running_threads = []
                missing_threads = copy.copy(self.threads)
                # we are done if all threads have finished
                while len(missing_threads) > 0:
                    # we spawn more threads whenever some threads have finished
                    if len(running_threads) < maxthreads:
                        to_index = min(
                            waiting_index + maxthreads - len(running_threads),
                            len(self.threads))
                        for i in range(waiting_index, to_index):
                            self.threads[i].start()
                            running_threads.append(self.threads[i])
                        waiting_index = to_index
                    # threads have exitcode only once they terminated
                    missing_threads = [t for t in missing_threads if t.exitcode is None]
                    running_threads = [t for t in running_threads if t.exitcode is None]
                    if (not self.silent
                        and len(running_threads) > 0):

                        print("[%s] still active"%",".join([th.worker.element.get_local_name() for th in running_threads]))
                    for thread in running_threads:
                        # this should prevent busy waiting
                        thread.join(1)
            except KeyboardInterrupt as k:
                for thread in self.threads:
                    if thread is not None and thread.is_alive():
                        print("[%s] terminated while active"%thread.worker.element.get_local_name())
                        thread.terminate()
                raise k

        self.outputs = [x for x in self.outputs if x is not None]
        message = ''
        for output in self.outputs:
            if "error" in output:
                if 'entry' in output:
                    message += "Error processing '%s' : %s\n"%(output['entry'].get_local_name(), output["error"])
                else:
                    message += "%s\n"%output["error"]
        if message != '':
            raise MultiProjectException(message)
        return self.outputs



# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import os
import shutil
import datetime

from vcstools import VcsClient
from vcstools.vcs_base import VcsError

from rosinstall.common import samefile, MultiProjectException
from rosinstall.config_yaml import PathSpec
from rosinstall.ui import Ui


# helper class
class PreparationReport(object):
    """
    Specifies after user interaction of how to perform install / update
    """
    
    def __init__(self, element):
        self.config_element = element
        self.abort = False       # abort ALL operations
        self.skip = False        # skip this tree
        self.error = None        # message
        self.verbose = False     # verbosity
        self.checkout = True     # checkout vs update
        self.backup = False      # backup vs delete
        self.backup_path = None  # where to move tree to

        
## Each Config element provides actions on a local folder
class ConfigElement(object):
    """
    Base class for Config provides methods with not implemented
    exceptions. Also a few shared methods.
    """

    def __init__(self, path, local_name, properties=None):
        self.path = path
        if path is None:
            raise MultiProjectException("Invalid empty path")
        self.local_name = local_name
        self.properties = properties

    def get_path(self):
        """A normalized absolute path"""
        return self.path

    def get_local_name(self):
        """What the user specified in his config"""
        return self.local_name

    def prepare_install(self, backup_path=None, arg_mode='abort', robust=False):
        """
        Check whether install can be performed, asking user for
        decision if necessary.

        :param arg_mode: one of prompt, backup, delete, skip.
        Determines how to handle error cases
        :param backup_path: if arg_mode==backup, determines where to backup to
        :param robust: if true, operation will be aborted without
        changes to the filesystem and without user interaction
        :returns: A preparation_report instance,
        telling whether to checkout or to update,
        how to deal with existing tree, and where to backup to.
        """
        preparation_report = PreparationReport(self)
        preparation_report.skip = True
        return preparation_report

    def install(self, checkout=True, backup=False, backup_path=None, verbose=False):
        """
        Attempt to make it so that self.path is the result of checking
        out / updating from remote repo.

        No user Interaction allowed here (for concurrent mode).

        :param checkout: whether to checkout or update
        :param backup: if checking out, what to do if path exists.
        If true, backup_path must be set.
        """
        raise NotImplementedError("ConfigElement install unimplemented")

    def get_path_spec(self):
        """PathSpec object with values as specified in file"""
        raise NotImplementedError("ConfigElement get_path_spec unimplemented")

    def get_properties(self):
        """Any meta information attached"""
        return self.properties

    def get_versioned_path_spec(self):
        """PathSpec where VCS elements have the version looked up"""
        raise NotImplementedError("ConfigElement get_versioned_path_spec unimplemented")

    def is_vcs_element(self):
        # subclasses to override when appropriate
        return False

    def get_diff(self, basepath=None):
        raise NotImplementedError("ConfigElement get_diff unimplemented")

    def get_status(self, basepath=None, untracked=False):
        raise NotImplementedError("ConfigElement get_status unimplemented")

    def backup(self, backup_path):
        if not backup_path:
            raise MultiProjectException(
                "[%s] Cannot install %s.  backup disabled."%(self.get_local_name(),
                                                             self.get_path()))
        backup_path = os.path.join(
            backup_path,
            "%s_%s"%(os.path.basename(self.path),
                     datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")))
        print("[%s] Backing up %s to %s"%(self.get_local_name(),
                                          self.get_path(),
                                          backup_path))
        shutil.move(self.path, backup_path)

    def __str__(self):
        return str(self.get_path_spec().get_legacy_yaml())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.get_path_spec() == other.get_path_spec()
        else:
            return False


class OtherConfigElement(ConfigElement):

    def __init__(self, path, local_name, uri=None, version='', properties=None):
        super(OtherConfigElement, self).__init__(path, local_name, properties)
        self.uri = uri
        self.version = version

    def install(self, checkout=True, backup=False, backup_path=None, verbose=False):
        return True

    def get_versioned_path_spec(self):
        raise MultiProjectException(
            "Cannot generate versioned outputs with non source types")

    def get_path_spec(self):
        "yaml as from source"
        version = self.version
        if version == '': version = None
        return PathSpec(local_name=self.get_local_name(),
                        path=self.get_path(),
                        scmtype=None,
                        uri=self.uri,
                        version=version,
                        tags=self.get_properties())

    def get_diff(self, basepath=None):
        return ''

    def get_status(self, basepath=None, untracked=False):
        return ''


class SetupConfigElement(ConfigElement):
    """
    A setup config element specifies a single file containing
    configuration data for a config.
    """

    def install(self, checkout=True, backup=False, backup_path=None, verbose=False):
        return True

    def get_versioned_path_spec(self):
        raise MultiProjectException(
            "Cannot generate versioned outputs with non source types")

    def get_path_spec(self):
        return PathSpec(local_name=self.get_local_name(),
                        path=self.get_path(),
                        tags=['setup-file'] + (self.get_properties() or []))

    def get_diff(self, basepath=None):
        return ''

    def get_status(self, basepath=None, untracked=False):
        return ''


class VCSConfigElement(ConfigElement):

    def __init__(self, path, local_name, uri, version='', properties=None):
        """
        Creates a config element for a VCS repository.

        :param path: absolute or relative path, str
        :param vcs_client: Object compatible with vcstools.VcsClientBase
        :param local_name: display name for the element, str
        :param uri: VCS uri to checkout/pull from, str
        :param version: optional revision spec (tagname, SHAID, ..., str)
        """
        super(VCSConfigElement, self).__init__(path, local_name, properties)
        if uri is None:
            raise MultiProjectException(
                "Invalid scm entry having no uri attribute for path %s"%path)
        # strip trailing slashes if defined to not be too strict #3061
        self.uri = uri.rstrip('/')
        self.version = version

    def _get_vcsc(self):
        raise NotImplementedError("VCSConfigElement _get_vcsc() unimplemented")

    def is_vcs_element(self):
        return True

    def get_vcs_type_name(self):
        # also see override in AVCSConfigElement
        return self._get_vcsc().get_vcs_type_name()

    def detect_presence(self):
        # also see override in AVCSConfigElement
        return self._get_vcsc().detect_presence()

    def path_exists(self):
        # we could use _get_vcsc().path_exit(), but this causes a time
        # penalty for initializing this is crucial for bash tab
        # completion
        return os.path.isdir(self.path)

    def prepare_install(self, backup_path=None, arg_mode='abort', robust=False):
        preparation_report = PreparationReport(self)
        present = self.detect_presence()
        if present or self.path_exists():
            # Directory exists see what we need to do
            error_message = None

            if not present:
                error_message = "Failed to detect %s presence at %s."%(
                    self.get_vcs_type_name(), self.path)
            else:
                cur_url = self._get_vcsc().get_url()
                if cur_url is not None:
                    cur_url = cur_url.rstrip('/')  #strip trailing slashes for #3269
                if not cur_url or cur_url != self.uri.rstrip('/'):
                    # local repositories get absolute pathnames
                    if not (os.path.isdir(self.uri) and
                            os.path.isdir(cur_url) and
                            samefile(cur_url, self.uri)):
                        if not self._get_vcsc().url_matches(cur_url, self.uri):
                            error_message = "Url %s does not match %s requested."%(
                                cur_url, self.uri)
            if error_message is None:
                # update should be possible
                preparation_report.checkout = False
            else:
                # If robust ala continue-on-error, just error now and
                # it will be continued at a higher level
                if robust:
                    raise MultiProjectException("Update Failed of %s: %s"%(
                            self.path, error_message))
                # prompt the user based on the error code
                if arg_mode == 'prompt':
                    print("Prepare updating %s (version %s) to %s"%(
                            self.uri, self.version, self.path))
                    mode = Ui.get_ui().prompt_del_abort_retry(error_message,
                                                              allow_skip=True)
                else:
                    mode = arg_mode
                if mode == 'backup':
                    preparation_report.backup = True
                    if backup_path is None:
                        print("Prepare updating %s (version %s) to %s"%(
                                self.uri, self.version, self.path))
                        preparation_report.backup_path = Ui.get_ui().get_backup_path()
                    else:
                        preparation_report.backup_path = backup_path
                if mode == 'abort':
                    preparation_report.abort = True
                    preparation_report.error = error_message
                if mode == 'skip':
                    preparation_report.skip = True
                    preparation_report.error = error_message
                if mode == 'delete':
                    preparation_report.backup = False
        return preparation_report

    def install(self, checkout=True, backup=True, backup_path=None, verbose=False):
        """
        Runs the equivalent of SCM checkout for new local repos or
        update for existing.

        :param checkout: whether to use an update command or
        a checkout/clone command
        :param backup: if checkout is True and folder exists,
        if backup is false folder will be DELETED.
        :param backup_path: if checkout is true and backup is true,
        move folder to this location
        """
        if checkout is True:
            print("[%s] Installing %s (version %s) to %s"%(
                    self.get_local_name(), self.uri, self.version, self.get_path()))
            if self.path_exists():
                if (backup is False):
                    shutil.rmtree(self.path)
                else:
                    self.backup(backup_path)
            if not self._get_vcsc().checkout(self.uri,
                                             self.version,
                                             verbose=verbose):
                raise MultiProjectException(
                    "[%s] Checkout of %s version %s into %s failed."%(
                        self.get_local_name(),
                        self.uri,
                        self.version,
                        self.get_path()))
        else:
            print("[%s] Updating %s"%(self.get_local_name(), self.get_path()))
            if not self._get_vcsc().update(self.version, verbose=verbose):
                raise MultiProjectException(
                    "[%s] Update Failed of %s"%(self.get_local_name(),
                                                self.get_path()))
        print("[%s] Done."%self.get_local_name())

    def get_path_spec(self):
        "yaml as from source"
        version = self.version
        if version == '':
            version = None
        return PathSpec(local_name=self.get_local_name(),
                        path=self.get_path(),
                        scmtype=self.get_vcs_type_name(),
                        uri=self.uri,
                        version=version,
                        tags=self.get_properties())

    def get_versioned_path_spec(self):
        "yaml looking up current version"
        version = self.version
        if version == '':
            version = None
        revision = None
        if version is not None:
            # revision is the UID of the version spec, can be them same
            revision = self._get_vcsc().get_version(self.version)
        currevision = self._get_vcsc().get_version()
        uri = self.uri
        curr_uri = self._get_vcsc().get_url()
        # uri might be a shorthand notation equivalent to curr_uri
        if self._get_vcsc().url_matches(curr_uri, uri):
            curr_uri = uri
        return PathSpec(local_name=self.get_local_name(),
                        path=self.get_path(),
                        scmtype=self.get_vcs_type_name(),
                        uri=self.uri,
                        version=version,
                        revision=revision,
                        currevision=currevision,
                        curr_uri=curr_uri,
                        tags=self.get_properties())


    def get_diff(self, basepath=None):
        return self._get_vcsc().get_diff(basepath)

    def get_status(self, basepath=None, untracked=False):
        return self._get_vcsc().get_status(basepath, untracked)



class AVCSConfigElement(VCSConfigElement):
    """
    Implementation using vcstools vcsclient, works for types svn, git, hg, bzr, tar

    :raises: Lookup Exception for unknown types
    """
    def __init__(self, scmtype, path, local_name, uri,
                 version='', vcsc=None, properties=None):
        super(AVCSConfigElement, self).__init__(path,
                                                local_name=local_name,
                                                uri=uri,
                                                version=version,
                                                properties=properties)
        self.vcsc = vcsc
        self._scmtype = scmtype

    def get_vcs_type_name(self):
        return self._scmtype

    def _get_vcsc(self):
        # lazy initializer
        if self.vcsc is None:
            try:
                self.vcsc = VcsClient(self._scmtype, self.get_path())
            except VcsError as e:
                raise MultiProjectException(
                    "Unable to create vcs client of type %s for %s: %s"%(
                        self._scmtype, self.get_path(), e))
        return self.vcsc

    def detect_presence(self):
        # to make more use of lazy initializer, do not instantiate
        # client for just this this is crucial for bash tab completion
        if self.get_vcs_type_name() == 'git':
            return os.path.isdir(os.path.join(self.path, '.git'))
        elif self.get_vcs_type_name() == 'svn':
            return os.path.isdir(os.path.join(self.path, '.svn'))
        elif self.get_vcs_type_name() == 'hg':
            return os.path.isdir(os.path.join(self.path, '.hg'))
        elif self.get_vcs_type_name() == 'bzr':
            return os.path.isdir(os.path.join(self.path, '.bzr'))
        else:
            return self._get_vcsc().detect_presence()



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""%(prog)s is a command to manipulate ROS workspaces. %(prog)s replaces its predecessor rosinstall.

Official usage:
  %(prog)s CMD [ARGS] [OPTIONS]

rosws will try to infer install path from context

Type '%(prog)s help' for usage.
"""


import os
import sys
import yaml
import shutil

from optparse import OptionParser

from rosinstall.cli_common import get_info_list, get_info_table, get_workspace
import rosinstall.rosinstall_cmd as rosinstall_cmd
from rosinstall.multiproject_cmd import get_config, cmd_install_or_update,\
    cmd_snapshot, cmd_version, cmd_info
import rosinstall.__version__

from rosinstall.common import MultiProjectException, select_elements
from rosinstall.helpers import ROSInstallException, ROSINSTALL_FILENAME,\
    get_ros_package_path, get_ros_stack_path
from rosinstall.multiproject_cli import MultiprojectCLI, __MULTIPRO_CMD_DICT__,\
    IndentedHelpFormatterWithNL
from rosinstall.config_yaml import get_path_spec_from_yaml

## This file adds or extends commands from multiproject_cli where ROS
## specific output has to be generated.

# extend the commands of multiproject
__ROSWS_CMD_DICT__ = {
      "regenerate": "create ROS workspace specific setup files"}
__ROSWS_CMD_ALIASES__ = {'update': 'up',
                         'remove': 'rm',
                         'status': 'st',
                         'diff':   'di'}
__ROSWS_CMD_DICT__.update(__MULTIPRO_CMD_DICT__)


class RoswsCLI(MultiprojectCLI):

    def __init__(self, config_filename=ROSINSTALL_FILENAME):
        MultiprojectCLI.__init__(self,
                                 config_filename,
                                 rosinstall_cmd.cmd_persist_config)

    def cmd_init(self, argv):
        if self.config_filename is None:
            print('Error: Bug: config filename required for init')
            return 1
        parser = OptionParser(usage="""usage: rosws init [TARGET_PATH [SOURCE_PATH]]?""",
                              formatter=IndentedHelpFormatterWithNL(),
                              description=__MULTIPRO_CMD_DICT__["init"] + """

rosws init does the following:
  1. Reads folder/file/web-uri SOURCE_PATH looking for a rosinstall yaml
  2. Creates new %s file at TARGET-PATH
  3. Generates ROS setup files

SOURCE_PATH can e.g. be a folder like /opt/ros/electric
If PATH is not given, uses current dir.

Examples:
$ rosws init ~/fuerte /opt/ros/fuerte
"""%self.config_filename,
                              epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        parser.add_option("-c", "--catkin", dest="catkin", default=False,
                          help="Declare this is a catkin build.",
                          action="store_true")
        parser.add_option("--cmake-prefix-path", dest="catkinpp", default=None,
                          help="Where to set the CMAKE_PREFIX_PATH",
                          action="store")
        parser.add_option("--continue-on-error", dest="robust", default=False,
                          help="Continue despite checkout errors",
                          action="store_true")
        parser.add_option("-j", "--parallel", dest="jobs", default=1,
                          help="How many parallel threads to use for installing",
                          action="store")
        (options, args) = parser.parse_args(argv)
        if len(args) < 1:
            target_path = '.'
        else:
            target_path = args[0]

        if not os.path.isdir(target_path):
            if not os.path.exists(target_path):
                os.mkdir(target_path)
            else:
                print('Error: Cannot create in target path %s '%target_path)

        if os.path.exists(os.path.join(target_path, self.config_filename)):
            print('Error: There already is a workspace config file %s at "%s". Use rosws install/modify.'%(self.config_filename, target_path))
            return 1
        if len(args) > 2:
            parser.error('Too many arguments')
        config_uris = []
        if len(args) == 2:
            config_uris.append(args[1])
        if len(config_uris) > 0:
            print('Using ROS_ROOT: %s'%config_uris[0])

        config = get_config(basepath=target_path,
                            additional_uris=config_uris,
                            config_filename=self.config_filename)

        # includes ROS specific files

        print("Writing %s"%os.path.join(config.get_base_path(), self.config_filename))
        rosinstall_cmd.cmd_persist_config(config)

        ## install or update each element
        install_success = cmd_install_or_update(
            config,
            robust=False,
            num_threads=int(options.jobs))

        rosinstall_cmd.cmd_generate_ros_files(config,
                                              target_path,
                                              nobuild=True,
                                              rosdep_yes=False,
                                              catkin=options.catkin,
                                              catkinpp=options.catkinpp,
                                              no_ros_allowed=True)

        if not install_success:
            print("Warning: installation encountered errors, but --continue-on-error was requested.  Look above for warnings.")
        print("\nrosinstall update complete.")
        if (options.catkin is False
            and options.catkinpp is None):
            print("\nType 'source %s/setup.bash' to change into this environment. Add that source command to the bottom of your ~/.bashrc to set it up every time you log in.\n\nIf you are not using bash please see http://www.ros.org/wiki/rosinstall/NonBashShells "%os.path.abspath(target_path))
        return 0

    def cmd_merge(self, target_path, argv, config=None):
        parser = OptionParser(usage="usage: rosws merge [URI] [OPTIONS]",
                              formatter=IndentedHelpFormatterWithNL(),
                              description=__MULTIPRO_CMD_DICT__["merge"] + """.

The command merges config with given other rosinstall element sets, from files or web uris.

The default workspace will be inferred from context, you can specify one using -t.

By default, when an element in an additional URI has the same
local-name as an existing element, the existing element will be
replaced. In order to ensure the ordering of elements is as
provided in the URI, use the option --merge-kill-append.

Examples:
$ rosws merge someother.rosinstall

You can use '-' to pipe in input, as an example:
$ roslocate info robot_mode | rosws merge -
""",
                              epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        # same options as for multiproject
        parser.add_option("-a", "--merge-kill-append", dest="merge_kill_append",
                          default=False,
                          help="merge by deleting given entry and appending new one",
                          action="store_true")
        parser.add_option("-k", "--merge-keep", dest="merge_keep",
                          default=False,
                          help="merge by keeping existing entry and discarding new one",
                          action="store_true")
        parser.add_option("-r", "--merge-replace", dest="merge_replace",
                          default=True,
                          help="(default) merge by replacing given entry with new one maintaining ordering",
                          action="store_true")
        parser.add_option("-y", "--confirm-all", dest="confirm_all",
                          default='',
                          help="do not ask for confirmation unless strictly necessary",
                          action="store_true")
        # required here but used one layer above
        parser.add_option("-t", "--target-workspace", dest="workspace", default=None,
                          help="which workspace to use",
                          action="store")
        (options, args) = parser.parse_args(argv)

        if len(args) > 1:
            print("Error: Too many arguments.")
            print(parser.usage)
            return -1
        if len(args) == 0:
            print("Error: Too few arguments.")
            print(parser.usage)
            return -1

        config_uris = args

        specs = []
        if config_uris[0] == '-':
            pipedata = "".join(sys.stdin.readlines())
            try:
                yamldicts = yaml.load(pipedata)
            except yaml.YAMLError as e:
                raise MultiProjectException(
                    "Invalid yaml format: \n%s \n%s"%(pipedata, e))
            if yamldicts is None:
                parser.error("No Input read from stdin")
            options.confirm_all = True # cant have user interaction and piped input
            specs.extend([get_path_spec_from_yaml(x) for x in yamldicts])
            config_uris = []

        merge_strategy = None
        count_mergeoptions = 0
        if options.merge_kill_append:
            merge_strategy = 'KillAppend'
            count_mergeoptions += 1
        if options.merge_keep:
            merge_strategy = 'MergeKeep'
            count_mergeoptions += 1
        if options.merge_replace:
            merge_strategy = 'MergeReplace'
            count_mergeoptions += 1
        if count_mergeoptions > 1:
            parser.error("You can only provide one merge-strategy")
        # default option
        if count_mergeoptions == 0:
            merge_strategy = 'MergeReplace'

        (newconfig, path_changed) = self.prompt_merge(
            target_path,
            additional_uris=config_uris,
            additional_specs=specs,
            path_change_message="ROS_PACKAGE_PATH order changed",
            merge_strategy=merge_strategy,
            confirmed=options.confirm_all,
            config=config)
        if newconfig is not None:
            print("Config changed, maybe you need run rosws update to update SCM entries.")
            print("Overwriting %s"%os.path.join(newconfig.get_base_path(), self.config_filename))
            shutil.move(os.path.join(newconfig.get_base_path(), self.config_filename), "%s.bak"%os.path.join(newconfig.get_base_path(), self.config_filename))
            rosinstall_cmd.cmd_persist_config(newconfig)
            rosinstall_cmd.cmd_maybe_refresh_ros_files(newconfig)
            print("\nrosws update complete.")
            if path_changed:
                print("\nDo not forget to do ...\n$ source %s/setup.sh\n... in every open terminal."%target_path)
        else:
            print("Merge caused no change, matching elements already exist")
        return 0



    def cmd_regenerate(self, target_path, argv, config=None):
        parser = OptionParser(usage="usage: rosws regenerate",
                              formatter=IndentedHelpFormatterWithNL(),
                              description=__MULTIPRO_CMD_DICT__["remove"] + """

this command without options generates files setup.sh, setup.bash and
setup.zsh. Note that doing this is unnecessary in general, as these
files do not change anymore, unless you change from one ROS distro to
another (which you should never do like this, create a separate new
workspace instead), or you deleted or modified any of those files
accidentally.
""",
                              epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        parser.add_option("-c", "--catkin", dest="catkin", default=False,
                          help="Declare this is a catkin build.",
                          action="store_true")
        parser.add_option("--cmake-prefix-path", dest="catkinpp", default=None,
                          help="Where to set the CMAKE_PREFIX_PATH",
                          action="store")
        (options, args) = parser.parse_args(argv)
        if len(args) > 0:
            print("Error: Too many arguments.")
            print(parser.usage)
            return -1

        if config is None:
            config = get_config(
                target_path,
                additional_uris=[],
                config_filename=self.config_filename)
        elif config.get_base_path() != target_path:
            raise MultiProjectException(
                "Config path does not match %s %s "%(config.get_base_path(),
                                                     target_path))
        rosinstall_cmd.cmd_generate_ros_files(config,
                                              target_path,
                                              nobuild=True,
                                              rosdep_yes=False,
                                              catkin=options.catkin,
                                              catkinpp=options.catkinpp,
                                              no_ros_allowed=True)
        return 0


    def cmd_info(self, target_path, argv, reverse=True, config=None):
        only_option_valid_attrs = ['path', 'localname', 'version', 'revision', 'cur_revision', 'uri', 'cur_uri', 'scmtype']
        parser = OptionParser(usage="usage: rosws info [localname]* [OPTIONS]",
                              formatter=IndentedHelpFormatterWithNL(),
                              description=__MULTIPRO_CMD_DICT__["info"] + """

The Status (S) column shows
 x  for missing
 L  for uncommited (local) changes
 V  for difference in version and/or remote URI

The 'Version-Spec' column shows what tag, branch or revision was given
in the .rosinstall file. The 'UID' column shows the unique ID of the
current (and specified) version. The 'URI' column shows the configured
URL of the repo.

If status is V, the difference between what was specified and what is
real is shown in the respective column. For SVN entries, the url is
split up according to standard layout (trunk/tags/branches).  The
ROS_PACKAGE_PATH follows the order of the table, earlier entries
overlay later entries.

When given one localname, just show the data of one element in list form.
This also has the generic properties element which is usually empty.

The --only option accepts keywords: %s

Examples:
$ rosws info -t ~/ros/fuerte
$ rosws info robot_model
$ rosws info --yaml
$ rosws info --only=path,cur_uri,cur_revision robot_model geometry
"""%only_option_valid_attrs,
                              epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        parser.add_option("--data-only", dest="data_only", default=False,
                          help="Does not provide explanations",
                          action="store_true")
        parser.add_option("--no-pkg-path", dest="no_pkg_path", default=False,
                          help="Suppress ROS_PACKAGE_PATH.",
                          action="store_true")
        parser.add_option("--pkg-path-only", dest="pkg_path_only", default=False,
                          help="Shows only ROS_PACKAGE_PATH separated by ':'. Supercedes all other options.",
                          action="store_true")
        parser.add_option("--only", dest="only", default=False,
                          help="Shows comma-separated lists of only given comma-separated attribute(s).",
                          action="store")
        parser.add_option("--yaml", dest="yaml", default=False,
                          help="Shows only version of single entry. Intended for scripting.",
                          action="store_true")

        # -t option required here for help but used one layer above, see cli_common
        parser.add_option("-t", "--target-workspace", dest="workspace", default=None,
                          help="which workspace to use",
                          action="store")
        (options, args) = parser.parse_args(argv)

        if config is None:
            config = get_config(
                target_path,
                additional_uris=[],
                config_filename=self.config_filename)
        elif config.get_base_path() != target_path:
            raise MultiProjectException("Config path does not match %s %s "%(config.get_base_path(), target_path))
        if args == []:
            args = None
        # relevant for code completion, so these should yield quick response:
        if options.pkg_path_only:
            print(":".join(get_ros_package_path(config)))
            return 0
        elif options.only:
            only_options = options.only.split(",")
            if only_options == '':
                parser.error('No valid options given')

            lookup_required = False
            for attr in only_options:
                if not attr in only_option_valid_attrs:
                    parser.error("Invalid --only option '%s', valids are %s"%(attr, only_option_valid_attrs))
                if attr in ['cur_revision', 'cur_uri', 'revision']:
                    lookup_required = True
            elements = select_elements(config, args)
            for element in elements:
                if lookup_required and element.is_vcs_element():
                    spec = element.get_versioned_path_spec()
                else:
                    spec = element.get_path_spec()
                output = []
                for attr in only_options:
                    if 'localname' == attr:
                        output.append(spec.get_local_name() or '')
                    if 'path' == attr:
                        output.append(spec.get_path() or '')
                    if 'scmtype' == attr:
                        output.append(spec.get_scmtype() or '')
                    if 'uri' == attr:
                        output.append(spec.get_uri() or '')
                    if 'version' == attr:
                        output.append(spec.get_version() or '')
                    if 'revision' == attr:
                        output.append(spec.get_revision() or '')
                    if 'cur_uri' == attr:
                        output.append(spec.get_curr_uri() or '')
                    if 'cur_revision' == attr:
                        output.append(spec.get_current_revision() or '')
                print(','.join(output))
            return 0
        if options.yaml:
            source_aggregate = cmd_snapshot(config, localnames=args)
            print(yaml.safe_dump(source_aggregate))
            return 0

        # this call takes long, as it invokes scms.
        outputs = cmd_info(config, localnames=args)
        if args is not None and len(outputs) == 1:
            print(get_info_list(config.get_base_path(),
                                outputs[0],
                                options.data_only))
            return 0

        header = 'workspace: %s\nROS_ROOT: %s'%(target_path,
                                                get_ros_stack_path(config))
        print(header)
        if not options.no_pkg_path:
            table = get_info_table(config.get_base_path(),
                                   outputs,
                                   options.data_only,
                                   reverse=reverse)
            if table is not None and table != '':
                print("\n%s"%table)

        return 0


def usage():
    """
    Prints usage from file header and from dictionary, sorting entries
    """
    dvars = {'prog': 'rosws'}
    dvars.update(vars())
    print(__doc__%dvars)
    # using None to generate empty lines
    keys = ['help', 'init',
            None, 'set', 'merge', 'remove',
            None, 'update',
            None, 'info', 'status', 'diff',
            None, 'regenerate']
    for key in keys:
        if key in __ROSWS_CMD_ALIASES__:
            alias = ' (%s)'%__ROSWS_CMD_ALIASES__[key]
        else:
            alias = ''
        if key is not None:
            print(("%s%s"%(key, alias)).ljust(10) + '   \t' + __ROSWS_CMD_DICT__[key])
        else:
            print('')

def rosws_main(argv=None):
    """
    Calls the function corresponding to the first argument.
    """
    if argv is None:
        argv = sys.argv
    if (sys.argv[0] == '-c'):
        sys.argv = ['rosws'] + sys.argv[1:]
    if '--version' in argv:
        print("rosws: \t%s\n%s"%(rosinstall.__version__.version, cmd_version()))
        sys.exit(0)
    workspace = None
    if len(argv) < 2:
        try:
            workspace = get_workspace(argv,
                                      os.getcwd(),
                                      config_filename=ROSINSTALL_FILENAME,
                                      varname="ROS_WORKSPACE")
            argv.append('info')
        except MultiProjectException as e:
            print(str(e))
            usage()
            return 0

    if '--help' == argv[1]:
        usage()
        return 0

    try:
        command = argv[1]
        args = argv[2:]

        if command == 'help':
            if len(argv) < 3:
                usage()
                return 0

            else:
                command = argv[2]
                args = argv[3:]
                args.insert(0, "--help")
                # help help
                if command == 'help':
                    usage()
                    return 0
        cli = RoswsCLI()

        # commands for which we do not infer target workspace
        commands = {'init': cli.cmd_init}
        # commands which work on a workspace
        ws_commands = {
            'info':       cli.cmd_info,
            'remove':     cli.cmd_remove,
            'regenerate': cli.cmd_regenerate,
            'set':        cli.cmd_set,
            'merge':      cli.cmd_merge,
            'diff':       cli.cmd_diff,
            'status':     cli.cmd_status,
            'update':     cli.cmd_update}
        for label in list(ws_commands.keys()):
            if label in __ROSWS_CMD_ALIASES__:
                ws_commands[__ROSWS_CMD_ALIASES__[label]] = ws_commands[label]

        if command not in commands and command not in ws_commands:
            if os.path.exists(command):
                args = ['-t', command] + args
                command = 'info'
            else:
                if command.startswith('-'):
                    print("First argument must be name of a command: %s"%command)
                else:
                    print("Error: unknown command: %s"%command)
                usage()
                return 1

        if command in commands:
            return commands[command](args)
        else:
            if workspace is None and not '--help' in args and not '-h' in args:
                workspace = get_workspace(args,
                                          os.getcwd(),
                                          config_filename=ROSINSTALL_FILENAME,
                                          varname="ROS_WORKSPACE")
            return ws_commands[command](workspace, args)

    except KeyboardInterrupt:
        pass
    except ROSInstallException as e:
        sys.stderr.write("ERROR in rosinstall: %s\n"%str(e))
        return 1
    except MultiProjectException as e:
        sys.stderr.write("ERROR: %s\n"%str(e))
        return 1




# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import subprocess
from rosinstall.config_elements import SetupConfigElement

ROSINSTALL_FILENAME = ".rosinstall"


class ROSInstallException(Exception):
    pass


def is_path_stack(path):
    """

    @return: True if the path provided is the root of a stack.
    """
    stack_path = os.path.join(path, 'stack.xml')
    if os.path.isfile(stack_path):
        return True
    return False


def is_path_ros(path):
    """
    warning: exits with code 1 if stack document is invalid
    @param path: path of directory to check
    @type  path: str
    @return: True if path points to the ROS stack
    @rtype: bool
    """
    if path is None:
        return False
    if os.path.basename(path) == 'ros':
        stack_path = os.path.join(path, 'stack.xml')
        return os.path.isfile(stack_path)
    return False


def get_ros_root_from_setupfile(path):
    """ Return the ROS_ROOT if the path is a setup.sh file with an
    env.sh next to it which sets the ROS_ROOT

    :returns: path to ROS_ROOT or None
    """
    # For groovy, we rely on setup.sh setting ROS_ROOT, as no more rosbuild stack 'ros' exists
    dirpath, basename = os.path.split(path)
    if basename != 'setup.sh':
        return None

    # env.sh exists since fuerte
    setupfilename = os.path.join(dirpath, 'env.sh')
    if not os.path.isfile(setupfilename):
        return None

    cmd = "%s sh -c 'echo $ROS_ROOT'" % setupfilename
    local_env = os.environ
    if 'ROS_ROOT' in local_env:
        local_env.pop('ROS_ROOT')
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=local_env, shell=True)
    out = process.communicate()[0]

    return out.strip()


def get_ros_stack_path(config):
    """ Detect valid ROS_ROOT directories from the config elements"""

    found_paths = set()
    sources = {}
    for t in config.get_config_elements():
        if is_path_ros(t.get_path()):
            found_paths.add(t.get_path())
            sources[t.get_path()] = t.get_path()
            continue
        if isinstance(t, SetupConfigElement):
            ros_root = get_ros_root_from_setupfile(t.get_local_name())
            if ros_root:
                found_paths.add(ros_root)
                sources[t.get_local_name()] = ros_root
                continue

    if len(found_paths) > 1:
        raise ROSInstallException(
            "Multiple ros stacks found in config %s, Please elimate all but one.\nThey come from the following sources: %s\n" % (found_paths, sources))
    elif len(found_paths) == 1:
        return found_paths.pop()
    return None


def get_ros_package_path(config):
    """ Return the simplifed ROS_PACKAGE_PATH """
    code_trees = []
    for t in reversed(config.get_config_elements()):
        if not is_path_ros(t.get_path()):
            if not os.path.isfile(t.get_path()):
                code_trees.append(t.get_path())
    return code_trees



# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import yaml
try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen
    
from rosinstall.common import MultiProjectException

__REPOTYPES__ = ['svn', 'bzr', 'hg', 'git', 'tar']
__ALLTYPES__ = __REPOTYPES__ + ['other', 'setup-file']

## The Path spec is a leightweigt object to transport the
## specification of a config element between functions,
## independently of yaml structure.
## Specifications are persisted in yaml, this file deals
## with manipulations of any such structures representing configs as
## yaml.
## get_path_spec_from_yaml turns yaml into path_spec, and pathspec
## get_legacy_yaml returns yaml.


def get_yaml_from_uri(uri):
    """reads and parses yaml from a local file or remote uri"""
    stream = None
    try:
        try:
            if os.path.isfile(uri):
                stream = open(uri, 'r')
            else:
                stream = urlopen(uri)
        except IOError as ioe:
            raise MultiProjectException(
                "Is not a local file, nor able to download as a URL [%s]: %s\n" % (uri, ioe))
        except ValueError as vae:
            raise MultiProjectException(
                "Is not a local file, nor a valid URL [%s] : %s\n" % (uri, vae))

        if not stream:
            raise MultiProjectException("couldn't load config uri %s\n" % uri)
        try:
            yamldata = yaml.load(stream)
        except yaml.YAMLError as yame:
            raise MultiProjectException(
                "Invalid multiproject yaml format in [%s]: %s\n" % (uri, yame))

        # we want a list or a dict, but pyyaml parses xml as string
        if type(yamldata) == 'str':
            raise MultiProjectException(
                "Invalid multiproject yaml format in [%s]: %s\n" % (uri, yamldata))
    finally:
        if stream is not None:
            stream.close()
    return yamldata


def get_path_specs_from_uri(uri, config_filename=None, as_is=False):
    """
    Builds a list of PathSpec elements from several types of input locations, "uris".
    The function treats other workspace folders/files as special uris to prevent mutual conflicts.

    :param uri: a folder, a file, or a web url
    :param config_filename: name for files to be treated special as other workspaces
    :param as_is: do not rewrite, used for loading the current workspace config without rewriting
    """
    if os.path.isdir(uri):
        if (config_filename is not None and
            os.path.isfile(os.path.join(uri, config_filename))):

            uri = os.path.join(uri, config_filename)
        else:
            # plain folders returned as themselves
            return [PathSpec(local_name=uri)]
    yaml_spec = get_yaml_from_uri(uri)
    if yaml_spec is None:
        return []
    specs = [get_path_spec_from_yaml(x) for x in yaml_spec]

    if (config_filename is not None and
        not as_is and
        os.path.isfile(uri) and
        os.path.basename(uri) == config_filename):

        # treat config files and folders with such files special
        # to prevent 2 workspaces from interacting
        return rewrite_included_source(specs, os.path.dirname(uri))
    return specs


def rewrite_included_source(source_path_specs, source_dir):
    """
    assumes source_path_specs is the contents of a config file in
    another directory source dir. It rewrites all elements, by changing
    any relative path relative to source dir and changing vcs
    types to non-vcs types types, to prevent two environments from
    conflicting
    """
    for index, pathspec in enumerate(source_path_specs):
        local_name = os.path.normpath(os.path.join(source_dir,
                                                   pathspec.get_local_name()))
        pathspec.set_local_name(local_name)
        if pathspec.get_path() is not None:
            path = os.path.normpath(os.path.join(source_dir, pathspec.get_path()))
            pathspec.set_path(path)
        pathspec.detach_vcs_info()
        source_path_specs[index] = pathspec
    return source_path_specs


def aggregate_from_uris(config_uris, config_filename=None):
    """
    Builds a List of PathSpec from a list of location strings (uri,
    paths). If locations is a folder, attempts to find config_filename in it,
    and use "folder/config_filename" instead(rewriting element path and
    stripping scm nature), else add folder as PathSpec.    Anything else,
    parse yaml at location, and add a PathSpec for each element.

    :param config_uris: source of yaml
    :param config_filename: file to use when given a folder
    """
    aggregate_source_yaml = []
    # build up a merged list of config elements from all given config_uris
    if config_uris is not None:
        for loop_uri in config_uris:
            source_path_specs = get_path_specs_from_uri(loop_uri, config_filename)
            # deal with duplicates in Config class
            if source_path_specs is not None:
                assert type(source_path_specs) == list, 'Probable bug: expected list, was %s'%type(source_path_specs)
                aggregate_source_yaml.extend(source_path_specs)
    return aggregate_source_yaml


class PathSpec:
    def __init__(self,
                 # localname is used as ID, currently also is used as path
                 local_name,
                 scmtype=None,
                 uri=None,
                 version=None,
                 tags=None,
                 revision=None,
                 currevision=None,
                 path=None,
                 curr_uri=None):
        """Fills in local properties based on dict, unifies different syntaxes"""
        self._local_name = local_name
        self._path = path
        self._uri = uri
        self._curr_uri = curr_uri
        self._version = version
        self._scmtype = scmtype
        self._tags = tags or []
        self._revision = revision
        self._currevision = currevision

    def __str__(self):
        return str(self.get_legacy_yaml())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def detach_vcs_info(self):
        """if wrapper has VCS information, remove it to make it a plain folder"""
        if self._scmtype is not None:
            self._scmtype = None
            self._uri = None
            self._version = None
            self._revision = None
            self._currevision = None

    def get_legacy_type(self):
        """return one of __ALLTYPES__"""
        if self._scmtype is not None:
            return self._scmtype
        elif self._tags is not None and 'setup-file' in self._tags:
            return 'setup-file'
        return 'other'

    def get_legacy_yaml(self):
        """
        return something like
        {hg: {local-name: common,
              version: common-1.0.2,
              uri: https://kforge.org/common/}}
        """
        # TODO switch to new syntax
        properties = {'local-name' : self._local_name}
        if self._uri is not None:
            properties['uri'] = self._uri
        if self._version is not None:
            properties['version'] = self._version
        if self._tags is not None:
            for tag in self._tags:
                if tag != 'setup-file' and tag != []:
                    if type(tag) == dict:
                        properties.update(tag)
                    else:
                        properties[tag] = None
        yaml_dict = {self.get_legacy_type(): properties}
        return yaml_dict

    def get_local_name(self):
        return self._local_name

    def set_local_name(self, local_name):
        self._local_name = local_name

    def get_path(self):
        return self._path

    def set_path(self, path):
        self._path = path

    def get_tags(self):
        return self._tags

    def get_scmtype(self):
        return self._scmtype

    def get_version(self):
        return self._version

    def get_revision(self):
        return self._revision

    def get_current_revision(self):
        return self._currevision

    def get_uri(self):
        return self._uri

    def get_curr_uri(self):
        return self._curr_uri


def get_path_spec_from_yaml(yaml_dict):
    """
    Fills in local properties based on dict, unifies different syntaxes
    """
    local_name = None
    uri = None
    version = None
    scmtype = None
    tags = []
    if type(yaml_dict) != dict:
        raise MultiProjectException(
            "Yaml for each element must be in YAML dict form")
    # old syntax:
# - hg: {local-name: common_rosdeps,
#        version: common_rosdeps-1.0.2,
#        uri: https://kforge.ros.org/common/rosdepcore}
# - setup-file: {local-name: /opt/ros/fuerte/setup.sh}
# - other: {local-name: /opt/ros/fuerte/share/ros}
# - other: {local-name: /opt/ros/fuerte/share}
# - other: {local-name: /opt/ros/fuerte/stacks}
    if yaml_dict is None or len(yaml_dict) == 0:
        raise MultiProjectException("no element in yaml dict.")
    if len(yaml_dict) > 1:
        raise MultiProjectException(
            "too many keys in element dict %s" % (list(yaml_dict.keys())))
    if not list(yaml_dict.keys())[0] in __ALLTYPES__:
        raise MultiProjectException(
            "Unknown element type '%s'" % (list(yaml_dict.keys())[0]))
    firstkey = list(yaml_dict.keys())[0]
    if firstkey in __REPOTYPES__:
        scmtype = list(yaml_dict.keys())[0]
    if firstkey == 'setup-file':
        tags.append('setup-file')
    values = yaml_dict[firstkey]
    if values is not None:
        for key, value in list(values.items()):
            if key == "local-name":
                local_name = value
            elif key == "meta":
                tags.append({key: value})
            elif key == "uri":
                uri = value
            elif key == "version":
                version = value
            else:
                raise MultiProjectException(
                    "Unknown key %s in %s" % (key, yaml_dict))
    # global validation
    if local_name is None:
        raise MultiProjectException(
            "Config element without a local-name: %s" % (yaml_dict))
    if scmtype != None:
        if uri is None:
            raise MultiProjectException(
                "scm type without declared uri in %s" % (values))
    # local_name is fixed, path may be normalized, made absolute, etc.
    path = local_name
    return PathSpec(local_name=local_name,
                    path=path,
                    scmtype=scmtype,
                    uri=uri,
                    version=version,
                    tags=tags)

def generate_config_yaml(config, filename, header):
    """
    Writes file filename with header first and then the config as yaml
    """
    if not os.path.exists(config.get_base_path()):
        os.makedirs(config.get_base_path())
    with open(os.path.join(config.get_base_path(), filename), 'w+b') as f:
        if header is not None:
            f.write(header.encode('UTF-8'))
        content = yaml.safe_dump([x.get_legacy_yaml() for x in config.get_source()])
        f.write(content.encode('UTF-8'))




# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.



import os

from rosinstall.helpers import ROSInstallException, get_ros_stack_path

CATKIN_CMAKE_TOPLEVEL = """#
#  TOPLEVEL cmakelists
#
cmake_minimum_required(VERSION 2.8)
cmake_policy(SET CMP0003 NEW)
cmake_policy(SET CMP0011 NEW)

set(CMAKE_CXX_FLAGS_INIT "-Wall")

enable_testing()

include(${CMAKE_SOURCE_DIR}/workspace-config.cmake OPTIONAL)

list(APPEND CMAKE_PREFIX_PATH ${CMAKE_BINARY_DIR} ${CMAKE_BINARY_DIR}/cmake)

file(MAKE_DIRECTORY ${CMAKE_BINARY_DIR}/lib)
file(MAKE_DIRECTORY ${CMAKE_BINARY_DIR}/bin)

if (IS_DIRECTORY ${CMAKE_SOURCE_DIR}/catkin)
  message(STATUS "+++ catkin")
  set(CATKIN_BUILD_PROJECTS "ALL" CACHE STRING
    "List of projects to build, or ALL for all.  Use to completely exclude certain projects from cmake traversal.")
  add_subdirectory(catkin)
else()
  find_package(catkin)
endif()

catkin_workspace()
"""


def generate_catkin_cmake(path, catkinpp):
    with open(os.path.join(path, "CMakeLists.txt"), 'w') as cmake_file:
        cmake_file.write(CATKIN_CMAKE_TOPLEVEL)

    if catkinpp:
        with open(os.path.join(path, "workspace-config.cmake"), 'w') as config_file:
            config_file.write("set (CMAKE_PREFIX_PATH %s)"%catkinpp)

            
def generate_embedded_python():
    return """import sys, os, yaml;

filename = '.rosinstall'
workspace_path = os.path.abspath('.')

if 'ROS_WORKSPACE' in os.environ:
  workspace_path = os.environ['ROS_WORKSPACE']

filename = os.path.join(workspace_path, filename)

if not os.path.isfile(filename):
    sys.exit("There is no file at %s"%filename)

with open(filename, "r") as f:
  try:
    v = f.read();
  except Exception as e:
    sys.exit("Failed to read file: %s %s "%(filename, str(e)))

try:
  y = yaml.load(v);
except Exception as e:
  sys.exit("Invalid yaml in %s: %s "%(filename, str(e)))

if y is not None:

  # put all non-setupfile entries into ROS_PACKAGE_PATH
  paths = []
  for vdict in y:
    for k, v in vdict.items():
      if v is not None and k != "setup-file":
        path = os.path.join(workspace_path, v['local-name'])
        if not os.path.isfile(path):
          # add absolute path from workspace to relative paths
          paths.append(os.path.normpath(path))
        else:
          sys.stderr.write("ERROR: referenced path is a file, not a folder: %s"%path)
  output = ''
  # add paths in reverse order
  if len(paths) > 0:
    output += ':'.join(reversed(paths))

  # We also want to return the location of any setupfile elements
  output += 'ROSINSTALL_PATH_SETUPFILE_SEPARATOR'
  setupfile_paths = []
  for vdict in y:
    for k, v in vdict.items():
      if v is not None and k == "setup-file":
        path = os.path.join(workspace_path, v['local-name'])
        if not os.path.exists(path):
          sys.stderr.write("WARNING: referenced setupfile does not exist: %s"%path)
        elif os.path.isfile(path):
          setupfile_paths.append(path)
        else:
          sys.stderr.write("ERROR: referenced setupfile is a folder: %s"%path)
  output += ':'.join(setupfile_paths)

  # printing will store the result in the variable
  print(output)"""
    
        
def generate_setup_sh_text(workspacepath):

    pycode = generate_embedded_python()

    # overlay or standard
    text = """#!/usr/bin/env sh
# THIS IS A FILE AUTO-GENERATED BY rosinstall
# IT IS UNLIKELY YOU WANT TO EDIT THIS FILE BY HAND
# IF YOU WANT TO CHANGE THE ROS ENVIRONMENT VARIABLES
# USE THE rosinstall OR rosws TOOL INSTEAD.
# see: http://www.ros.org/wiki/rosinstall
"""

    # Sadly we cannot infer the workspacepath from within the sourced file
    text += """\nexport ROS_WORKSPACE=%s
if [ ! "$ROS_MASTER_URI" ] ; then export ROS_MASTER_URI=http://localhost:11311 ; fi
unset ROS_ROOT
"""%workspacepath

    # use python script to read ros_package_path and setup-file elements
    text += """
# python script to read .rosinstall even when rosinstall is not installed
# this files parses the .rosinstall and sets environment variables accordingly
# The ROS_PACKAGE_PATH contains all elements in reversed order (for historic reasons)

# We store into _PARSED_CONFIG the result of python code,
# which is the ros_package_path and the list of setup_files to source
# Using python here to benefit of the pyyaml library
export _PARSED_CONFIG=`/usr/bin/env python << EOPYTHON

%s
EOPYTHON`

#whitespace separates ros_package_path and setupfile results, using sed to split them up
_ROS_PACKAGE_PATH_ROSINSTALL=`echo "$_PARSED_CONFIG" | sed 's,\(.*\)ROSINSTALL_PATH_SETUPFILE_SEPARATOR\(.*\),\\1,'`
_SETUPFILES_ROSINSTALL=`echo "$_PARSED_CONFIG" | sed 's,\(.*\)'ROSINSTALL_PATH_SETUPFILE_SEPARATOR'\(.*\),\\2,'`
unset _PARSED_CONFIG

# colon separates entries
_LOOP_SETUP_FILE=`echo $_SETUPFILES_ROSINSTALL | sed 's,\([^:]*\)[:]\(.*\),\\1,'`
while [ ! -z "$_LOOP_SETUP_FILE" ]
do
  if [ -f "$_LOOP_SETUP_FILE" ]; then
    . $_LOOP_SETUP_FILE
  else
    echo warn: no such file : "$_LOOP_SETUP_FILE"
  fi
  _SETUPFILES_ROSINSTALL=`echo $_SETUPFILES_ROSINSTALL | sed 's,\([^:]*[:]*\),,'`
  _LOOP_SETUP_FILE=`echo $_SETUPFILES_ROSINSTALL | sed 's,\([^:]*\)[:]\(.*\),\\1,'`
done

unset _LOOP_SETUP_FILE
unset _SETUPFILES_ROSINSTALL

export ROS_PACKAGE_PATH=$_ROS_PACKAGE_PATH_ROSINSTALL
unset _ROS_PACKAGE_PATH_ROSINSTALL

# using ROS_ROOT now being in ROS_PACKAGE_PATH
export _ROS_ROOT_ROSINSTALL=`/usr/bin/env python << EOPYTHON
import sys, os;
if 'ROS_PACKAGE_PATH' in os.environ:
  pkg_path = os.environ['ROS_PACKAGE_PATH']
  for path in pkg_path.split(':'):
    if (os.path.basename(path) == 'ros'
        and os.path.isfile(os.path.join(path, 'stack.xml'))):
      print(path)
      break
EOPYTHON`

if [ ! -z "${_ROS_ROOT_ROSINSTALL}" ]; then
  export ROS_ROOT=$_ROS_ROOT_ROSINSTALL
  export PATH=$ROS_ROOT/bin:$PATH
  export PYTHONPATH=$ROS_ROOT/core/roslib/src:$PYTHONPATH
fi
unset _ROS_ROOT_ROSINSTALL
"""%pycode

    return text


def generate_setup_bash_text(shell, no_ros=False):
    '''
    Generates the contents that go into a setup.bash or setup.zsh
    file.  The intent of such a file is to enable shell extensions,
    such as special ros commands and tab completion.  The generation
    is complex because the setup of the system changed between ROS
    electric and fuerte. In fuerte, the distro setup.sh also loads
    distro rosbash based on CATKIN_SHELL. Before fuerte, it is up to
    setup.bash to do so.
    '''
    if shell == 'bash':
        script_path = """
SCRIPT_PATH="${BASH_SOURCE[0]}";
if([ -h "${SCRIPT_PATH}" ]) then
  while([ -h "${SCRIPT_PATH}" ]) do SCRIPT_PATH=`readlink "${SCRIPT_PATH}"`; done
fi
export OLDPWDBAK=$OLDPWD
pushd . > /dev/null
cd `dirname ${SCRIPT_PATH}` > /dev/null
SCRIPT_PATH=`pwd`;
popd  > /dev/null
export OLDPWD=$OLDPWDBAK
"""
    elif shell == 'zsh':
        script_path = 'SCRIPT_PATH="$(dirname $0)"'
    else:
        raise ROSInstallException("%s shell unsupported."%shell)

    text = """#!/usr/bin/env %(shell)s
# THIS IS A FILE AUTO-GENERATED BY rosinstall
# IT IS UNLIKELY YOU WANT TO EDIT THIS FILE BY HAND
# IF YOU WANT TO CHANGE THE ROS ENVIRONMENT VARIABLES
# USE THE rosinstall TOOL INSTEAD.
# see: http://www.ros.org/wiki/rosinstall

CATKIN_SHELL=%(shell)s

%(script_path)s

# Load the path of this particular setup.%(shell)s

if [ ! -f "$SCRIPT_PATH/setup.sh" ]; then
  echo "Bug: shell script unable to determine its own location: $SCRIPT_PATH"
  return 22
fi"""%locals()

    if not no_ros:
        text += """
# unset _ros_decode_path to check later whether setup.sh has sourced ros%(shell)s
unset -f _ros_decode_path 1> /dev/null 2>&1"""%locals()

        text += """
. $SCRIPT_PATH/setup.sh"""

    if not no_ros:
        text += """
# Cannot rely on $? due to set -o errexit in build scripts
RETURNCODE=`type _ros_decode_path 2> /dev/null | grep function 1>/dev/null 2>&1 || echo error`

if [ ! "$RETURNCODE" = "" ]; then
  RETURNCODE=`rospack help 1> /dev/null 2>&1 || echo error`
  if  [ "$RETURNCODE" = "" ]; then
    ROSSHELL_PATH=`rospack find rosbash`/ros%(shell)s
    if [ -e "$ROSSHELL_PATH" ]; then
      . $ROSSHELL_PATH
    fi
  else
    echo "rospack could not be found, you cannot have ros%(shell)s features until you bootstrap ros"
  fi
fi
"""%locals()
    return text


def generate_setup(config, no_ros_allowed=False):
    ros_root = get_ros_stack_path(config)
    if ros_root is None:
        if not no_ros_allowed:
            candidates = []
            for t in config.get_config_elements():
                if os.path.basename(t.get_local_name()) == 'ros':
                    candidates.append(t.get_path())
            raise ROSInstallException("""
No 'ros' stack detected in candidates %s.
Please add the location of a ros distribution to this command.

See http://ros.org/wiki/rosinstall."""%(candidates))
  
    text = generate_setup_sh_text(workspacepath=config.get_base_path())
    setup_path = os.path.join(config.get_base_path(), 'setup.sh')
    with open(setup_path, 'w') as f:
        f.write(text)

    for shell in ['bash', 'zsh']:
        text = generate_setup_bash_text(shell, ros_root is None)
        setup_path = os.path.join(config.get_base_path(), 'setup.%s'%shell)
        with open(setup_path, 'w') as f:
            f.write(text)



# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys

class Ui(object):
    """
    wrap user interaction, such that client libraries may provide own
    implementation
    """

    # For now, primarily define this for replacement in unittests
    GLOBAL_UI = None

    @staticmethod
    def get_ui():
        if Ui.GLOBAL_UI is None:
            return Ui()
        return Ui.GLOBAL_UI

    @staticmethod
    def set_ui(uiarg):
        Ui.GLOBAL_UI = uiarg

    def __init__(self):
        pass

    def get_backup_path(self):
        """Interactive function asking the user to choose a path for backup"""
        backup_path = self.get_input("Please enter backup pathname: ")
        print(("backing up to %s"%backup_path))
        return backup_path

    def get_input(self, prompt):
        if sys.hexversion > 0x03000000:
            return input(prompt)
        else:
            return raw_input(prompt)
    
    def prompt_del_abort_retry(self, prompt, allow_skip=False):
        """
        Interactive function asking the user to choose a conflict resolution
        """
        if allow_skip:
            valid_modes = ['(d)elete and replace',
                           '(a)bort',
                           '(b)ackup and replace',
                           '(s)kip']
        else:
            valid_modes = ['(d)elete and replace',
                           '(a)bort',
                           '(b)ackup and replace']

        mode = ""

        full_prompt = "%s %s: "%(prompt, ", ".join(valid_modes))

        while mode == "":
            mode_input = self.get_input(full_prompt)
            if mode_input == 'b' or mode_input == 'backup':
                mode = 'backup'
            elif mode_input == 'd' or mode_input == 'delete':
                mode = 'delete'
            elif mode_input == 'a' or mode_input == 'abort':
                mode = 'abort'
            elif allow_skip and mode_input == 's' or mode_input == 'skip':
                mode = 'skip'
        return mode



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
usage: rosinstall [OPTIONS] INSTALL_PATH [ROSINSTALL FILES OR DIRECTORIES]
see: http://www.ros.org/wiki/rosinstall

Common Option:
-n or --nobuild (don't perform a 'make core_cools' on the ros stack)

Type 'rosinstall --help' for usage.

Common invocations:

initial checkout:   rosinstall ~/ros "http://packages.ros.org/cgi-bin/gen_rosinstall.py?rosdistro=diamondback&variant=ros-full&overlay=no"
subsequent update:  rosinstall ~/ros

"""

from __future__ import print_function
import os
import sys
from optparse import OptionParser
import yaml
import shutil

from rosinstall import multiproject_cmd, rosinstall_cmd
from rosinstall.helpers import ROSINSTALL_FILENAME
import rosinstall.__version__

def usage():
    print(__doc__%vars())
    exit(1)


def rosinstall_main(argv):
    if len(argv) < 2:
        usage()
    args = argv[1:]
    parser = OptionParser(usage="usage: rosinstall [OPTIONS] INSTALL_PATH [ROSINSTALL FILES OR DIRECTORIES]\n\n\
rosinstall does the following:\n\
  1. Merges all URIs into new or existing .rosinstall file at PATH\n\
  2. Checks out or updates all version controlled URIs\n\
  3. If ros stack is installed from source, calls rosmake after checkout or updates.\n\
  4. Generates/overwrites updated setup files\n\n\
If running with --catkin mode:\
  1. Merges all URIs into new or existing .rosinstall file at PATH\n\
  2. Checks out or updates all version controlled URIs\n\
  4. Generates/overwrites updated setup files and creates CMakeLists.txt at the root.\n\n\
URIs can be web urls to remote .rosinstall files, local .rosinstall files,\n\
git, svn, bzr, hg URIs, or other (local directories)\n\
Later URIs will shadow packages of earlier URIs.\n",
                          epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
    parser.add_option("-c", "--catkin", dest="catkin", default=False,
                      help="Declare this is a catkin build.",
                      action="store_true")
    parser.add_option("--cmake-prefix-path", dest="catkinpp", default=None,
                      help="Where to set the CMAKE_PREFIX_PATH, implies --catkin",
                      action="store")
    parser.add_option("--version", dest="version", default=False,
                      help="display version information",
                      action="store_true")
    parser.add_option("--verbose", dest="verbose", default=False,
                      help="display more information",
                      action="store_true")
    parser.add_option("-n", "--nobuild", dest="nobuild", default=False,
                      help="skip the build step for the ROS stack",
                      action="store_true")
    parser.add_option("--rosdep-yes", dest="rosdep_yes", default=False,
                      help="Pass through --rosdep-yes to rosmake",
                      action="store_true")
    parser.add_option("--continue-on-error", dest="robust", default=False,
                      help="Continue despite checkout errors",
                      action="store_true")
    parser.add_option("--delete-changed-uris", dest="delete_changed",
                      default=False,
                      help="Delete the local copy of a directory before changing uri.",
                      action="store_true")
    parser.add_option("--abort-changed-uris", dest="abort_changed",
                      default=False,
                      help="Abort if changed uri detected",
                      action="store_true")
    parser.add_option("--backup-changed-uris", dest="backup_changed",
                      default='',
                      help="backup the local copy of a directory before changing uri to this directory.",
                      action="store")
    parser.add_option("--diff", dest="vcs_diff",
                      default=False,
                      help="shows a combined diff over all SCM entries",
                      action="store_true")
    parser.add_option("--status", dest="vcs_status",
                      default=False,
                      help="shows a combined status command over all SCM entries",
                      action="store_true")
    parser.add_option("--status-untracked", dest="vcs_status_untracked",
                      default=False,
                      help="shows a combined status command over all SCM entries, also showing untracked files",
                      action="store_true")
    parser.add_option("-j", "--parallel", dest="jobs",
                      default=1,
                      help="How many parallel threads to use for installing",
                      action="store")
    parser.add_option("--generate-versioned-rosinstall", dest="generate_versioned",
                      default=None,
                      help="generate a versioned rosinstall file", action="store")
    (options, args) = parser.parse_args(args)

    #if options.rosdep_yes:
    #  parser.error("rosinstall no longer bootstraps the build, it will not call rosmake or pass it rosdep options")

    if options.version:
        print("rosinstall %s\n%s"%(rosinstall.__version__.version, multiproject_cmd.cmd_version()))
        sys.exit(0)

    if len(args) < 1:
        parser.error("rosinstall requires at least 1 argument")

    mode = 'prompt'
    if options.delete_changed:
        mode = 'delete'
    if options.abort_changed:
        if mode == 'delete':
            parser.error("delete-changed-uris is mutually exclusive with abort-changed-uris")
        mode = 'abort'
    if options.backup_changed != '':
        if mode == 'delete':
            parser.error("delete-changed-uris is mutually exclusive with backup-changed-uris")
        if mode == 'abort':
            parser.error("abort-changed-uris is mutually exclusive with backup-changed-uris")
        mode = 'backup'

    # Catkin must be enabled if catkinpp is set
    if options.catkinpp:
        options.catkin = True

    # Get the path to the rosinstall
    options.path = os.path.abspath(args[0])

    config_uris = args[1:]

    config = multiproject_cmd.get_config(basepath=options.path,
                                         additional_uris=config_uris,
                                         config_filename=ROSINSTALL_FILENAME)

    if options.generate_versioned:
        filename = os.path.abspath(options.generate_versioned)
        source_aggregate = multiproject_cmd.cmd_snapshot(config)
        with open(filename, 'w') as fh:
            fh.write(yaml.safe_dump(source_aggregate))
        print("Saved versioned rosinstall of current directory %s to %s"%(options.path, filename))
        return True

    if options.vcs_diff:
        difflist = multiproject_cmd.cmd_diff(config)
        alldiff = []
        for entrydiff in difflist:
            if entrydiff['diff'] is not None and entrydiff['diff'] != '':
                alldiff.append(entrydiff['diff'])
        print('\n'.join(alldiff))
        return True

    if options.vcs_status or options.vcs_status_untracked:
        statuslist = multiproject_cmd.cmd_status(
          config,
          untracked=options.vcs_status_untracked)
        allstatus = ""
        for entrystatus in statuslist:
            if entrystatus['status'] is not None:
                allstatus += entrystatus['status']
        print(allstatus, end='')
        return True

    print("rosinstall operating on", options.path,
                "from specifications in rosinstall files ",
                ", ".join(config_uris))

    # includes ROS specific files
    print("(Over-)Writing %s"%os.path.join(options.path, ROSINSTALL_FILENAME))
    if(os.path.isfile(os.path.join(options.path, ROSINSTALL_FILENAME))):
        shutil.move(os.path.join(options.path, ROSINSTALL_FILENAME),
                    "%s.bak"%os.path.join(options.path, ROSINSTALL_FILENAME))
    rosinstall_cmd.cmd_persist_config(config)

    ## install or update each element
    install_success = multiproject_cmd.cmd_install_or_update(
        config,
        backup_path=options.backup_changed,
        mode=mode,
        robust=options.robust,
        num_threads=int(options.jobs),
        verbose=options.verbose)

    rosinstall_cmd.cmd_generate_ros_files(
        config,
        options.path,
        options.nobuild,
        options.rosdep_yes,
        options.catkin,
        options.catkinpp)

    if not install_success:
        print("Warning: installation encountered errors, but --continue-on-error was requested.    Look above for warnings.")

    print("\nrosinstall update complete.")
    if (options.catkin is False
        and options.catkinpp is None):
      
        print("\n\nNow, type 'source %s/setup.bash' to set up your environment.\nAdd that to the bottom of your ~/.bashrc to set it up every time.\n\nIf you are not using bash please see http://www.ros.org/wiki/rosinstall/NonBashShells "%os.path.abspath(options.path))
    return True



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import os
import subprocess
from rosinstall.multiproject_cmd import cmd_persist_config as multipersist
from rosinstall import setupfiles
from rosinstall.helpers import ROSINSTALL_FILENAME, is_path_ros


def cmd_persist_config(config, config_filename=ROSINSTALL_FILENAME):
    ## Save .rosinstall
    header = """# THIS IS A FILE WHICH IS MODIFIED BY rosinstall
# IT IS UNLIKELY YOU WANT TO EDIT THIS FILE BY HAND,
# UNLESS FOR REMOVING ENTRIES.
# IF YOU WANT TO CHANGE THE ROS ENVIRONMENT VARIABLES
# USE THE rosinstall TOOL INSTEAD.
# IF YOU CHANGE IT, USE rosinstall FOR THE CHANGES TO TAKE EFFECT
"""
    multipersist(config, config_filename, header)


def _ros_requires_boostrap(config):
    """Whether we might need to bootstrap ros"""
    for entry in config.get_source():
        if is_path_ros(os.path.join(config.get_base_path(), entry.get_local_name())):
            # we assume that if any of the elements we installed came
            # from a VCS source, a bootsrap might be useful
            if entry.get_scmtype() is not None:
                return True
    return False


def cmd_maybe_refresh_ros_files(config):
    """
    Regenerates setup.* files if they exist already
    """
    if (os.path.isfile(os.path.join(config.get_base_path(), 'setup.sh'))):
        print("Overwriting setup.sh, setup.bash, and setup.zsh in %s"%config.get_base_path())
        setupfiles.generate_setup(config, no_ros_allowed=True)


def cmd_generate_ros_files(config, path, nobuild=False, rosdep_yes=False, catkin=False, catkinpp=None, no_ros_allowed=False):
    """
    Generates ROS specific setup files
    """

    # Catkin must be enabled if catkinpp is set
    if catkinpp is not None:
        catkin = True

    ## bootstrap the build if installing ros
    if catkin:
        setupfiles.generate_catkin_cmake(path, catkinpp)

    else: # DRY install case
        ## Generate setup.sh and save
        print("(Over-)Writing setup.sh, setup.bash, and setup.zsh in %s"%config.get_base_path())
        setupfiles.generate_setup(config, no_ros_allowed)

        if _ros_requires_boostrap(config) and not nobuild:
            print("Bootstrapping ROS build")
            rosdep_yes_insert = ""
            if rosdep_yes:
                rosdep_yes_insert = " --rosdep-yes"
            ros_comm_insert = ""
            if 'ros_comm' in [os.path.basename(tree.get_path()) for tree in config.get_config_elements()]:
                print("Detected ros_comm bootstrapping it too.")
                ros_comm_insert = " ros_comm"
            subprocess.check_call("source %s && rosmake ros%s --rosdep-install%s" % (os.path.join(path, 'setup.sh'), ros_comm_insert, rosdep_yes_insert), shell=True, executable='/bin/bash')






# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function
import os
import textwrap
import shutil
from optparse import OptionParser, IndentedHelpFormatter

from rosinstall.common import samefile, select_element, select_elements,\
    MultiProjectException, normalize_uri, string_diff
from rosinstall.config_yaml import PathSpec
import rosinstall.multiproject_cmd as multiproject_cmd
from rosinstall.ui import Ui

# implementation of single CLI commands (extracted for use in several overlapping scripts)

__MULTIPRO_CMD_DICT__ = {
    "help":     "provide help for commands",
    "init":     "set up a directory as workspace",
    "info":     "Overview of some entries",
    "merge":    "merges your workspace with another config set",
    "set":      "add or changes one entry from your workspace config",
    "update":   "update or check out some of your config elements",
    "remove":   "remove an entry from your workspace config, without deleting files",
    "snapshot": "write a file specifying repositories to have the version they currently have",
    "diff":     "print a diff over some SCM controlled entries",
    "status":   "print the change status of files in some SCM controlled entries"}


class IndentedHelpFormatterWithNL(IndentedHelpFormatter):
    def format_description(self, description):
        if not description: return ""
        desc_width = self.width - self.current_indent
        indent = " "*self.current_indent
        # the above is still the same
        bits = description.split('\n')
        formatted_bits = [
            textwrap.fill(bit,
                          desc_width,
                          initial_indent=indent,
                          subsequent_indent=indent)
            for bit in bits]
        result = "\n".join(formatted_bits) + "\n"
        return result


def _get_mode_from_options(parser, options):
    mode = 'prompt'
    if options.delete_changed:
        mode = 'delete'
    if options.abort_changed:
        if mode == 'delete':
            parser.error("delete-changed-uris is mutually exclusive with abort-changed-uris")
        mode = 'abort'
    if options.backup_changed != '':
        if mode == 'delete':
            parser.error("delete-changed-uris is mutually exclusive with backup-changed-uris")
        if mode == 'abort':
            parser.error("abort-changed-uris is mutually exclusive with backup-changed-uris")
        mode = 'backup'
    return mode



class MultiprojectCLI:

    def __init__(self,
                 config_filename=None,
                 config_generator=multiproject_cmd.cmd_persist_config):
        self.config_filename = config_filename
        self.config_generator = config_generator

    # def cmd_init(self, argv):
    #     raise Exception("Not implemented yet")
    # TODO enable when making multiproject an independent CLI


    # def cmd_merge(self, target_path, argv, config=None):
    #     raise Exception("Not implemented yet")
    # TODO enable when making multiproject an independent CLI


    def cmd_diff(self, target_path, argv, config=None):
        parser = OptionParser(usage="usage: rosws diff [localname]* ",
                              description=__MULTIPRO_CMD_DICT__["diff"],
                              epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        # required here but used one layer above
        parser.add_option("-t", "--target-workspace", dest="workspace",
                          default=None,
                          help="which workspace to use",
                          action="store")
        (_, args) = parser.parse_args(argv)

        if config is None:
            config = multiproject_cmd.get_config(
                target_path,
                additional_uris=[],
                config_filename=self.config_filename)
        elif config.get_base_path() != target_path:
            raise MultiProjectException(
                "Config path does not match %s %s "%(config.get_base_path(),
                                                     target_path))


        if len(args) > 0:
            difflist = multiproject_cmd.cmd_diff(config, localnames=args)
        else:
            difflist = multiproject_cmd.cmd_diff(config)
        alldiff = []
        for entrydiff in difflist:
            if entrydiff['diff'] is not None and entrydiff['diff'] != '':
                alldiff.append(entrydiff['diff'])
        print('\n'.join(alldiff))

        return False


    def cmd_status(self, target_path, argv, config=None):
        parser = OptionParser(usage="usage: rosws status [localname]* ",
                              description=__MULTIPRO_CMD_DICT__["status"] + ". The status columns meanings are as the respective SCM defines them.",
                              epilog="""See: http://www.ros.org/wiki/rosinstall for details""")
        parser.add_option("--untracked", dest="untracked",
                          default=False,
                          help="Also shows untracked files",
                          action="store_true")
        # -t option required here for help but used one layer above, see cli_common
        parser.add_option("-t", "--target-workspace", dest="workspace",
                          default=None,
                          help="which workspace to use",
                          action="store")
        (options, args) = parser.parse_args(argv)

        if config is None:
            config = multiproject_cmd.get_config(
                target_path,
                additional_uris=[],
                config_filename=self.config_filename)
        elif config.get_base_path() != target_path:
            raise MultiProjectException(
                "Config path does not match %s %s "%(config.get_base_path(),
                                                     target_path))

        if len(args) > 0:
            statuslist = multiproject_cmd.cmd_status(config,
                                                     localnames=args,
                                                     untracked=options.untracked)
        else:
            statuslist = multiproject_cmd.cmd_status(config,
                                                     untracked=options.untracked)
        allstatus = ""
        for entrystatus in statuslist:
            if entrystatus['status'] is not None:
                allstatus += entrystatus['status']
        print(allstatus, end='')
        return 0


    def cmd_set(self, target_path, argv, config=None):
        """
        command for modifying/adding a single entry
        :param target_path: where to look for config
        :param config: config to use instead of parsing file anew
        """
        parser = OptionParser(usage="usage: rosws set [localname] [SCM-URI]?  [--(detached|svn|hg|git|bzr)] [--version=VERSION]]",
                              formatter=IndentedHelpFormatterWithNL(),
                              description=__MULTIPRO_CMD_DICT__["set"] + """
The command will infer whether you want to add or modify an entry. If
you modify, it will only change the details you provide, keeping
those you did not provide. if you only provide a uri, will use the
basename of it as localname unless such an element already exists.

The command only changes the configuration, to checkout or update
the element, run rosws update afterwards.

Examples:
$ rosws set robot_model --hg https://kforge.ros.org/robotmodel/robot_model
$ rosws set robot_model --version robot_model-1.7.1
$ rosws set robot_model --detached
""",
                              epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        parser.add_option("--detached", dest="detach", default=False,
                          help="make an entry unmanaged (default for new element)",
                          action="store_true")
        parser.add_option("-v", "--version-new", dest="version", default=None,
                          help="point SCM to this version",
                          action="store")
        parser.add_option("--git", dest="git", default=False,
                          help="make an entry a git entry",
                          action="store_true")
        parser.add_option("--svn", dest="svn", default=False,
                          help="make an entry a subversion entry",
                          action="store_true")
        parser.add_option("--hg", dest="hg", default=False,
                          help="make an entry a mercurial entry",
                          action="store_true")
        parser.add_option("--bzr", dest="bzr", default=False,
                          help="make an entry a bazaar entry",
                          action="store_true")
        parser.add_option("-y", "--confirm", dest="confirm", default='',
                          help="Do not ask for confirmation",
                          action="store_true")
        # -t option required here for help but used one layer above, see cli_common
        parser.add_option("-t", "--target-workspace", dest="workspace", default=None,
                          help="which workspace to use",
                          action="store")
        (options, args) = parser.parse_args(argv)

        if len(args) > 2:
            print("Error: Too many arguments.")
            print(parser.usage)
            return -1

        if config is None:
            config = multiproject_cmd.get_config(
                target_path,
                additional_uris=[],
                config_filename=self.config_filename)
        elif config.get_base_path() != target_path:
            raise MultiProjectException(
                "Config path does not match %s %s "%(config.get_base_path(),
                                                     target_path))

        scmtype = None
        count_scms = 0
        if options.git:
            scmtype = 'git'
            count_scms += 1
        if options.svn:
            scmtype = 'svn'
            count_scms += 1
        if options.hg:
            scmtype = 'hg'
            count_scms += 1
        if options.bzr:
            scmtype = 'bzr'
            count_scms += 1
        if options.detach:
            count_scms += 1
        if count_scms > 1:
            parser.error("You cannot provide more than one scm provider option")

        if len(args) == 0:
            parser.error("Must provide a localname")

        element = select_element(config.get_config_elements(), args[0])

        uri = None
        if len(args) == 2:
            uri = args[1]
        version = None
        if options.version is not None:
            version = options.version.strip("'\"")

        # create spec object
        if element is None:
            # asssume is insert, choose localname
            localname = os.path.normpath(args[0])
            rel_path = os.path.relpath(os.path.realpath(localname),
                                       os.path.realpath(config.get_base_path()))
            if os.path.isabs(localname):
                # use shorter localname for folders inside workspace
                if not rel_path.startswith('..'):
                    localname = rel_path
            else:
                # got a relative path as localname, could point to a dir or be meant relative to workspace
                if not samefile(os.getcwd(), config.get_base_path()):
                    if os.path.isdir(localname):
                        parser.error(
                            "Cannot decide which one you want to add:\n%s\n%s"%(
                                os.path.abspath(localname),
                                os.path.join(config.get_base_path(), localname)))
                    if not rel_path.startswith('..'):
                        localname = rel_path

            spec = PathSpec(local_name=localname,
                            uri=normalize_uri(uri, config.get_base_path()),
                            version=version,
                            scmtype=scmtype)
        else:
            # modify
            old_spec = element.get_path_spec()
            if options.detach:
                spec = PathSpec(local_name=element.get_local_name())
            else:
                # '' evals to False, we do not want that
                if version is None:
                    version = old_spec.get_version()
                spec = PathSpec(local_name=element.get_local_name(),
                                uri=normalize_uri(uri or old_spec.get_uri(),
                                                  config.get_base_path()),
                                version=version,
                                scmtype=scmtype or old_spec.get_scmtype(),
                                path=old_spec.get_path())
            if spec.get_legacy_yaml() == old_spec.get_legacy_yaml():
                if not options.detach and spec.get_scmtype() is not None:
                    parser.error("Element %s already exists, did you mean --detached ?"%spec)
                parser.error("Element %s already exists"%spec)

        (newconfig, path_changed) = self.prompt_merge(
            target_path,
            additional_uris=[],
            additional_specs=[spec],
            merge_strategy='MergeReplace',
            confirmed=options.confirm,
            confirm=not options.confirm,
            show_verbosity=False,
            show_advanced=False,
            config=config)

        if newconfig is not None:
            print("Overwriting %s"%os.path.join(newconfig.get_base_path(), self.config_filename))
            shutil.move(os.path.join(newconfig.get_base_path(), self.config_filename), "%s.bak"%os.path.join(newconfig.get_base_path(), self.config_filename))
            self.config_generator(newconfig, self.config_filename)
            if path_changed:
                print("\nDo not forget to do ...\n$ source %s/setup.sh\n... in every open terminal."%target_path)
            if (spec.get_scmtype() is not None):
                print("Config changed, remember to run 'rosws update %s' to update the folder from %s"%(spec.get_local_name(), spec.get_scmtype()))
        else:
            print("New element %s could not be added, "%spec)
            return 1
        # auto-install not a good feature, maybe make an option
        # for element in config.get_config_elements():
        #   if element.get_local_name() == spec.get_local_name():
        #     if element.is_vcs_element():
        #       element.install(checkout=not os.path.exists(os.path.join(config.get_base_path(), spec.get_local_name())))
        #       break
        return 0


    def cmd_update(self, target_path, argv, config=None):
        parser = OptionParser(usage="usage: rosws update [localname]*",
                              formatter=IndentedHelpFormatterWithNL(),
                              description=__MULTIPRO_CMD_DICT__["update"] + """

This command calls the SCM provider to pull changes from remote to
your local filesystem. In case the url has changed, the command will
ask whether to delete or backup the folder.

Examples:
$ rosws update -t ~/fuerte
$ rosws update robot_model geometry
""",
                              epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        parser.add_option("--delete-changed-uris", dest="delete_changed",
                          default=False,
                          help="Delete the local copy of a directory before changing uri.",
                          action="store_true")
        parser.add_option("--abort-changed-uris", dest="abort_changed",
                          default=False,
                          help="Abort if changed uri detected",
                          action="store_true")
        parser.add_option("--continue-on-error", dest="robust",
                          default=False,
                          help="Continue despite checkout errors",
                          action="store_true")
        parser.add_option("--backup-changed-uris", dest="backup_changed",
                          default='',
                          help="backup the local copy of a directory before changing uri to this directory.",
                          action="store")
        parser.add_option("-j", "--parallel", dest="jobs",
                          default=1,
                          help="How many parallel threads to use for installing",
                          action="store")
        parser.add_option("-v", "--verbose", dest="verbose",
                          default=False,
                          help="Whether to print out more information",
                          action="store_true")
        # -t option required here for help but used one layer above, see cli_common
        parser.add_option("-t", "--target-workspace", dest="workspace",
                          default=None,
                          help="which workspace to use",
                          action="store")
        (options, args) = parser.parse_args(argv)

        if config is None:
            config = multiproject_cmd.get_config(
                target_path,
                additional_uris=[],
                config_filename=self.config_filename)
        elif config.get_base_path() != target_path:
            raise MultiProjectException("Config path does not match %s %s "%(
                    config.get_base_path(),
                    target_path))
        success = True
        mode = _get_mode_from_options(parser, options)
        if args == []:
            # None means no filter, [] means filter all
            args = None
        if success:
            install_success = multiproject_cmd.cmd_install_or_update(
                config,
                localnames=args,
                backup_path=options.backup_changed,
                mode=mode,
                robust=options.robust,
                num_threads=int(options.jobs),
                verbose=options.verbose)
            if install_success or options.robust:
                return 0
        return 1


    def cmd_remove(self, target_path, argv, config=None):
        parser = OptionParser(usage="usage: rosws remove [localname]*",
                              formatter=IndentedHelpFormatterWithNL(),
                              description=__MULTIPRO_CMD_DICT__["remove"] + """
The command removes entries from your configuration file, it does not affect your filesystem.
""",
                              epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        (_, args) = parser.parse_args(argv)
        if len(args) < 1:
            print("Error: Too few arguments.")
            print(parser.usage)
            return -1

        if config is None:
            config = multiproject_cmd.get_config(
                target_path,
                additional_uris=[],
                config_filename=self.config_filename)
        elif config.get_base_path() != target_path:
            raise MultiProjectException(
                "Config path does not match %s %s "%(config.get_base_path(),
                                                     target_path))
        success = True
        elements = select_elements(config, args)
        for element in elements:
            if not config.remove_element(element.get_local_name()):
                success = False
                print("Bug: No such element %s in config, aborting without changes"%(element.get_local_name()))
                break
        if success:
            print("Overwriting %s"%os.path.join(config.get_base_path(),
                                                self.config_filename))
            shutil.move(os.path.join(config.get_base_path(),
                                     self.config_filename),
                        "%s.bak"%os.path.join(config.get_base_path(),
                                              self.config_filename))
            self.config_generator(config, self.config_filename)
            print("Removed entries %s"%args)

        return 0

    # def cmd_info(self, target_path, argv, reverse=False, config=None):
    #     """
    #     :param target_path: where to look for config
    #     :param config: config to use instead of parsing file anew
    #     """
    #     __pychecker__ = 'unusednames=reverse'
    #     raise Exception("Not implemented yet")
    # TODO enable when making multiproject an independent CLI


    def _get_element_diff(self, new_path_spec, config_old, extra_verbose=False):
        """
        :returns: a string telling what changed for element compared to old config
        """
        if new_path_spec is None or config_old is None:
            return ''
        output = [' %s'%new_path_spec.get_local_name()]
        if extra_verbose:
            old_element = None
            if config_old is not None:
                old_element = select_element(config_old.get_config_elements(),
                                             new_path_spec.get_local_name())

            if old_element is None:
                if new_path_spec.get_scmtype() is not None:
                    output.append("   \t%s  %s   %s"%(new_path_spec.get_scmtype(),
                                                      new_path_spec.get_uri(),
                                                      new_path_spec.get_version() or ''))
            else:
                old_path_spec = old_element.get_path_spec()
                accessor_map = {PathSpec.get_scmtype: 'scmtype',
                                PathSpec.get_version: 'version',
                                PathSpec.get_revision: 'revision',
                                PathSpec.get_current_revision: 'current revision',
                                PathSpec.get_curr_uri: 'current_uri',
                                PathSpec.get_uri: 'specified uri'}
                for accessor, label in list(accessor_map.items()):
                    old_val = accessor(old_path_spec)
                    new_val = accessor(new_path_spec)
                    if old_val is not None and\
                            old_val != new_val:
                        diff = string_diff(old_val, new_val)
                        output.append("  \t%s: %s -> %s;"%(label, old_val, diff))
                    elif old_val is None and\
                            new_val is not None and\
                            new_val != "" and\
                            new_val != []:
                        output.append("  %s = %s"%(label,
                                                   new_val))
        return ''.join(output)


    def prompt_merge(self,
                     target_path,
                     additional_uris,
                     additional_specs,
                     path_change_message=None,
                     merge_strategy='KillAppend',
                     confirmed=False,
                     confirm=False,
                     show_advanced=True,
                     show_verbosity=True,
                     config=None):
        """
        Prompts the user for the resolution of a merge. Without
        further options, will prompt only if elements change. New
        elements are just added without prompt.

        :param target_path: Location of the config workspace
        :param additional_uris: uris from which to load more elements
        :param additional_specs: path specs for additional elements
        :param path_change_message: Something to tell the user about elements order
        :param merge_strategy: See Config.insert_element
        :param confirmed: Never ask
        :param confirm: Always ask, supercedes confirmed
        :param config: None or a Config object for target path if available
        :param show_advanced: if true allow to change merge strategy
        :param show_verbosity: if true allows to change verbosity
        :returns: tupel (Config or None if no change, bool path_changed)
        """
        if config is None:
            config = multiproject_cmd.get_config(
                target_path,
                additional_uris=[],
                config_filename=self.config_filename)
        elif config.get_base_path() != target_path:
            msg = "Config path does not match %s %s "%(config.get_base_path(),
                                                       target_path)
            raise MultiProjectException(msg)
        local_names_old = [x.get_local_name() for x in config.get_config_elements()]

        extra_verbose = confirmed or confirm
        abort = None
        last_merge_strategy = None
        while abort is None:

            if (last_merge_strategy is None
                or last_merge_strategy != merge_strategy):
                newconfig = multiproject_cmd.get_config(
                    target_path,
                    additional_uris=[],
                    config_filename=self.config_filename)
                config_actions = multiproject_cmd.add_uris(
                    newconfig,
                    additional_uris=additional_uris,
                    merge_strategy=merge_strategy)
                for path_spec in additional_specs:
                    action = newconfig.add_path_spec(path_spec, merge_strategy)
                    config_actions[path_spec.get_local_name()] = (action, path_spec)
                last_merge_strategy = merge_strategy

            local_names_new = [x.get_local_name() for x in newconfig.get_config_elements()]

            path_changed = False
            ask_user = False
            output = ""
            new_elements = []
            changed_elements = []
            discard_elements = []
            for localname, (action, new_path_spec) in list(config_actions.items()):
                index = -1
                if localname in local_names_old:
                    index = local_names_old.index(localname)
                if action == 'KillAppend':
                    ask_user = True
                    if (index > -1 and local_names_old[:index + 1] == local_names_new[:index + 1]):
                        action = 'MergeReplace'
                    else:
                        changed_elements.append(self._get_element_diff(new_path_spec, config, extra_verbose))
                        path_changed = True

                if action == 'Append':
                    path_changed = True
                    new_elements.append(self._get_element_diff(new_path_spec,
                                                               config,
                                                               extra_verbose))
                elif action == 'MergeReplace':
                    changed_elements.append(self._get_element_diff(new_path_spec,
                                                                   config,
                                                                   extra_verbose))
                    ask_user = True
                elif action == 'MergeKeep':
                    discard_elements.append(self._get_element_diff(new_path_spec,
                                                                   config,
                                                                   extra_verbose))
                    ask_user = True
            if len(changed_elements) > 0:
                output += "\n     Change details of element (Use --merge-keep or --merge-replace to change):\n"
                if extra_verbose:
                    output += " %s\n"%("\n".join(changed_elements))
                else:
                    output += " %s\n"%(", ".join(changed_elements))
            if len(new_elements) > 0:
                output += "\n     Add new elements:\n"
                if extra_verbose:
                    output += " %s\n"%("\n".join(new_elements))
                else:
                    output += " %s\n"%(", ".join(new_elements))


            if local_names_old != local_names_new[:len(local_names_old)]:
                old_order = ' '.join(reversed(local_names_old))
                new_order = ' '.join(reversed(local_names_new))
                output += "\n     %s (Use --merge-keep or --merge-replace to prevent) from\n %s\n     to\n %s\n\n"%(path_change_message or "Element order change", old_order, new_order)
                ask_user = True

            if output == "":
                return (None, False)
            if not confirm and (confirmed or not ask_user):
                print("     Performing actions: ")
                print(output)
                return (newconfig, path_changed)
            else:
                print(output)
                showhelp = True
                while(showhelp):
                    showhelp = False
                    prompt = "Continue: (y)es, (n)o"
                    if show_verbosity:
                        prompt +=", (v)erbosity"
                    if show_advanced:
                        prompt += ", (a)dvanced options"
                    prompt += ": "
                    mode_input = Ui.get_ui().get_input(prompt)
                    if mode_input == 'y':
                        return (newconfig, path_changed)
                    elif mode_input == 'n':
                        abort = True
                    elif show_advanced and mode_input == 'a':
                        strategies = {'MergeKeep': "(k)eep",
                                      'MergeReplace': "(s)witch in",
                                      'KillAppend': "(a)ppending"}
                        unselected = [v for k, v in list(strategies.items()) if k != merge_strategy]
                        print("""New entries will just be appended to the config and
appear at the beginning of your ROS_PACKAGE_PATH. The merge strategy
decides how to deal with entries having a duplicate localname or path.

"(k)eep" means the existing entry will stay as it is, the new one will
be discarded. Useful for getting additional elements from other
workspaces without affecting your setup.

"(s)witch in" means that the new entry will replace the old in the
same position. Useful for upgrading/downgrading.

"switch (a)ppend" means that the existing entry will be removed, and
the new entry appended to the end of the list. This maintains order
of elements in the order they were given.

Switch append is the default.
""")
                        prompt = """Change Strategy %s: """%(", ".join(unselected))
                        mode_input = Ui.get_ui().get_input(prompt)
                        if mode_input == 's':
                            merge_strategy = 'MergeReplace'
                        elif mode_input == 'k':
                            merge_strategy = 'MergeKeep'
                        elif mode_input == 'a':
                            merge_strategy = 'KillAppend'

                    elif show_verbosity and mode_input == 'v':
                        extra_verbose = not extra_verbose
            if abort:
                print("No changes made.")
            print('==========================================')
        return (None, False)



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import os
import re
from optparse import OptionParser
from rosinstall.common import samefile, MultiProjectException

# Support for any command line interface (CLI) for rosinstall


def get_workspace(argv, shell_path, config_filename=None, varname=None):
    """
    If target option -t is given return value of that one. Else, varname
    is given and exists, considers that one, plus,
    if config_filename is given, searches for a file named in config_filename
    in '.', above '.'.
    In that case, if two solutions are found, asks the user.

    :returns: abspath if a .rosinstall was found, error and exist else.
    """
    parser = OptionParser()
    parser.add_option("-t", "--target-workspace", dest="workspace", default=None,
                      help="which workspace to use",
                      action="store")
    # suppress errors based on any other option this parser is agnostic about
    argv2 = [x for x in argv if ((not x.startswith('-')) or
                              x.startswith('--target-workspace=') or
                              x.startswith('-t=') or
                              x in ['-t', '--target-workspace'])]
    (options, args) = parser.parse_args(argv2)
    if options.workspace is not None:
        if (config_filename is not None and
            not os.path.isfile(os.path.join(options.workspace, config_filename))):

            raise MultiProjectException("%s has no workspace configuration file '%s'"%(os.path.abspath(options.workspace), config_filename))
        return os.path.abspath(options.workspace)

    varname_path = None
    if varname is not None and varname in os.environ:
        # workspace could be relative, maybe confusing, but that's the users fault
        varname_path = os.environ[varname]
        if varname_path.strip() == '' or not os.path.isdir(varname_path):
            varname_path = None

    # use current dir
    current_path = None
    if config_filename is not None:
        while shell_path is not None and not shell_path == os.path.dirname(shell_path):
            if os.path.exists(os.path.join(shell_path, config_filename)):
                current_path = shell_path
                break
            shell_path = os.path.dirname(shell_path)

    if current_path is not None and varname_path is not None and not samefile(current_path, varname_path):
        raise MultiProjectException("Ambiguous workspace: %s=%s, %s"%(varname, varname_path, os.path.abspath(config_filename)))

    if current_path is None and varname_path is None:
        raise MultiProjectException("Command requires a target workspace.")

    if current_path is not None:
        return current_path
    else:
        return varname_path


def _uris_match(basepath, uri1, uri2):
    """
    True if uri2 is None or not None and same folder or equal string
    as uri1. Relative folders resolved using basepath
    """
    if uri1 is None:
        uri1 = ''
    if uri2 is None:
        return True
    if ((uri1 == uri2) or
        (basepath is not None and
         os.path.isdir(os.path.join(basepath, uri1)) and
         os.path.realpath(os.path.join(basepath, uri2)) == os.path.realpath(os.path.join(basepath, uri1)))):
        return True
    return False


def _get_status_flags(basepath, line):
    if 'exists' in line and line['exists'] is False:
        return 'x'
    mflag = ''
    if 'modified' in line and line['modified'] is True:
        mflag = 'M'
    if (('curr_uri' in line and
         not _uris_match(basepath, line['uri'], line['curr_uri'])) or
        ('specversion' in line and
         line['specversion'] is not None and
         line['actualversion'] is not None and
         line['specversion'] != line['actualversion'])):
        mflag += 'V'
    return mflag

def get_info_table_elements(basepath, entries, headers):
    """returns a list of dict with refined information from entries"""

    outputs = []
    for line in entries:
        if not 'curr_uri' in line:
            line['curr_uri'] = None
        if not 'specversion' in line:
            line['specversion'] = None
        if not 'actualversion' in line:
            line['actualversion'] = None
        if not 'curr_version' in line:
            line['curr_version'] = None
        if not 'version' in line:
            line['version'] = None
        output_dict = {'scm': line['scm'],
                       'uri': line['uri'],
                       'curr_uri': None,
                       'version': line['version'],
                       'localname': line['localname']}

        if line is None:
            print("Bug Warning, an element is missing")
            continue

        if line['scm'] == 'git':
            if (line['specversion'] is not None and len(line['specversion']) > 12):
                line['specversion'] = line['specversion'][0:12]
            if (line['actualversion'] is not None and len(line['actualversion']) > 12):
                line['actualversion'] = line['actualversion'][0:12]

        if line['scm'] is not None:

            if line['scm'] == 'svn':
                # in case of SVN, we can use the final part of standard uri as version
                uri = line['uri']
                version = line['version']
                m = re.match('(.*/)((tags|branches|trunk)(/.*)*)', uri)
                if m is not None and len(m.groups()) > 1 and uri == ''.join(m.groups()[0:2]):
                    uri = m.groups()[0]
                    if (m.groups()[1] is not None and version is None or version.strip() == ''):
                        version = m.groups()[1]
                    else:
                        version = m.groups()[1]
                    line['uri'] = uri
                    line['version'] = version
                if line['curr_uri'] is not None:
                    uri = line['curr_uri']
                    m = re.match('(.*/)((tags|branches|trunk)(/.*)*)', uri)
                    if m is not None and len(m.groups()) > 1 and uri == ''.join(m.groups()[0:2]):
                        uri = m.groups()[0]
                        if (m.groups()[1] is not None and version is None or version.strip() == ''):
                            version = m.groups()[1]
                        else:
                            version = m.groups()[1]
                        line['curr_uri'] = uri
                        line['curr_version'] = version

            if (line['curr_version'] is not None and
                line['version'] != line['curr_version']):

                output_dict['version'] = "%s  (%s)"%(line['curr_version'], line['version'])
            else:
                output_dict['version'] = line['version']

            if (line['specversion'] is not None and
                line['specversion'] != '' and
                line['actualversion'] != line['specversion']):

                output_dict['matching'] = "%s (%s)"%(line['actualversion'], line['specversion'])
            else:
                output_dict['matching'] = line['actualversion']

            common_prefixes = ["https://", "http://"]
            if line['uri'] is not None:
                for pre in common_prefixes:
                    if line['uri'].startswith(pre):
                        line['uri'] = line['uri'][len(pre):]
                        break
                output_dict['uri'] = line['uri']

            if line['curr_uri'] is not None:
                for pre in common_prefixes:
                    if line['curr_uri'].startswith(pre):
                        line['curr_uri'] = line['curr_uri'][len(pre):]
                        break

            if (not _uris_match(basepath, line['uri'], line['curr_uri'])):
                output_dict['uri'] = "%s    (%s)"%(line['curr_uri'], line['uri'])

        else:
            output_dict['matching'] = " "
        output_dict['status'] = _get_status_flags(basepath, line)

        outputs.append(output_dict)

    return outputs

def get_info_table(basepath, entries, data_only=False, reverse=False):
    """return a refined textual representation of the entries"""
    headers = {
      'uri':"URI  (Spec) [http(s)://...]",
      'scm':"SCM ",
      'localname':"Localname",
      'version':"Version-Spec",
      'matching': "UID  (Spec)",
      'status':"S"
      }

    # table design
    selected_headers = ['localname', 'status', 'scm', 'version', 'matching', 'uri']

    outputs = get_info_table_elements(
        basepath=basepath,
        entries=entries,
        headers=headers)

    # adjust column width
    column_length = {}
    for header in list(headers.keys()):
        column_length[header] = len(headers[header])
        for entry in outputs:
            if entry[header] is not None:
                column_length[header] = max(column_length[header],
                                            len(entry[header]))

    resultlines = []
    if not data_only and len(outputs) > 0:
        header_line = ' '
        for i, header in enumerate(selected_headers):
            output = headers[header]
            if i < len(selected_headers) - 1:
                output = output.ljust(column_length[header]) + " "
            header_line += output
        resultlines.append(header_line)
        header_line = ' '
        for i, header in enumerate(selected_headers):
            output = '-' * len(headers[header])
            if i < len(selected_headers) - 1:
                output = output.ljust(column_length[header]) + " "
            header_line += output
        resultlines.append(header_line)

    if reverse:
        outputs = reversed(outputs)
    for entry in outputs:
        if entry is None:
            print("Bug Warning, an element is missing")
            continue
        data_line = ' '
        for i, header in enumerate(selected_headers):
            output = entry[header]
            if output is None:
                output = ''
            if i < len(selected_headers) - 1:
                output = output.ljust(column_length[header]) + " "
            data_line += output
        resultlines.append(data_line)
    return "\n".join(resultlines)


def get_info_list(basepath, line, data_only=False):
    """
    Info for a single config entry
    """

    assert line is not None, "Bug Warning, an element is missing"

    headers = {
      'uri'           : "URI:",
      'curr_uri'      : "Current URI:",
      'scm'           : "SCM:",
      'localname'     : "Localname:",
      'path'          : "Path",
      'version'       : "Version-Spec:",
      'status'        : "Status:",
      'specversion'   : "Spec-Revision:",
      'actualversion' : "Current-Revision:",
      'properties'    : "Other Properties:"}

    # table design
    selected_headers = ['localname', 'path', 'status',
                        'scm', 'uri', 'curr_uri',
                        'version', 'specversion', 'actualversion',
                        'properties']

    line['status'] = _get_status_flags(basepath, line)

    header_length = 0
    for header in list(headers.keys()):
        header_length = max(header_length, len(headers[header]))
    result = ''
    for header in selected_headers:
        if not data_only:
            title = "%s  "%(headers[header].ljust(header_length))
        else:
            title = ''
        if header in line:
            output = line[header]
        if output is None:
            output = ''
        result += "%s%s\n"%(title, output)
    return result



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


"""
The _cmd python files attempt to provide a reasonably
complete level of abstraction to multiproject functionality.

Client code will need to pass the Config element through,
and may use the ConfigElement API in places.
There are no guarantees at this time for the API to
remain stable, but the cmd API probably will change least.
A change to expect is abstraction of user interaction.
"""


import os
from rosinstall.common import MultiProjectException, DistributedWork,\
    select_elements, normabspath
from rosinstall.config import Config, realpath_relation
from rosinstall.config_yaml import aggregate_from_uris, generate_config_yaml,\
    get_path_specs_from_uri

import vcstools
import vcstools.__version__


def get_config(basepath,
               additional_uris=None,
               config_filename=None,
               merge_strategy='KillAppend'):
    """
    Create a Config element necessary for all other commands.  The
    command will look at the uris in sequence, each can be a web
    resource, a filename or a folder. In case it is a folder, when a
    config_filename is provided, the folder will be searched for a
    file of that name, and that one will be used.  Else the folder
    will be considered a target location for the config.  All files
    will be parsed for config elements, thus conceptually the input to
    Config is an expanded list of config elements. Config takes this
    list and consolidates duplicate paths by keeping the last one in
    the list.

    :param basepath: where relative paths shall be resolved against
    :param additional_uris: the location of config specifications or folders
    :param config_filename: name of files which may be looked at for config information
    :param merge_strategy: One of 'KillAppend, 'MergeKeep', 'MergeReplace'
    :returns: a Config object
    :raises MultiProjectException: on plenty of errors
    """
    if basepath is None:
        raise MultiProjectException("Need to provide a basepath for Config.")

    #print("source...........................", path_specs)

    ## Generate the config class with the uri and path
    if (config_filename is not None
        and basepath is not None
        and os.path.isfile(os.path.join(basepath, config_filename))):

        base_path_specs = get_path_specs_from_uri(os.path.join(basepath,
                                                               config_filename),
                                                  as_is=True)
    else:
        base_path_specs = []

    config = Config(base_path_specs, basepath,
                    config_filename=config_filename,
                    merge_strategy=merge_strategy)

    add_uris(config, additional_uris, merge_strategy)

    return config


def add_uris(config, additional_uris, merge_strategy="KillAppend"):
    """
    changes the given config by merging with the additional_uris

    :param config: a Config objects
    :param additional_uris: the location of config specifications or folders
    :param config_filename: name of files which may be looked at for config information
    :param merge_strategy: One of 'KillAppend, 'MergeKeep', 'MergeReplace'
    :returns: a dict {<local-name>: (<action>, <path-spec>), <local-name>: ...} determined by the merge_strategy
    :raises MultiProjectException: on plenty of errors
    """
    if config is None:
        raise MultiProjectException("Need to provide a Config.")

    if additional_uris is None or len(additional_uris) == 0:
        return {}

    added_uris = []
    if config.get_config_filename() is not None:
        for uri in additional_uris:
            comp_uri = None
            if (os.path.isfile(uri)
                and os.path.basename(uri) == config.get_config_filename()):

                comp_uri = os.path.dirname(uri)
            if (os.path.isdir(uri)
                and os.path.isfile(os.path.join(uri, config.get_config_filename()))):
                comp_uri = uri
            if (comp_uri is not None and
                realpath_relation(os.path.abspath(comp_uri),
                                  os.path.abspath(config.get_base_path())) == 'SAME_AS'):
                print('Warning: Discarding config basepath as additional uri: %s'%uri)
                continue
            added_uris.append(uri)

    path_specs = aggregate_from_uris(added_uris, config.get_config_filename())

    actions = {}
    for path_spec in path_specs:
        action = config.add_path_spec(path_spec, merge_strategy)
        actions[path_spec.get_local_name()] = (action, path_spec)

    return actions


def cmd_persist_config(config, filename, header=None):
    """writes config to given file in yaml syntax"""
    generate_config_yaml(config, filename, header)


def cmd_version():
    """Returns extensive version information"""
    def prettyversion(vdict):
        version = vdict.pop("version")
        return "%s; %s"%(version, ",".join(list(vdict.values())))
    return """vcstools:  %s
SVN:       %s
Mercurial: %s
Git:       %s
Tar:       %s
Bzr:       %s
"""%(vcstools.__version__.version,
     prettyversion(vcstools.SvnClient.get_environment_metadata()),
     prettyversion(vcstools.HgClient.get_environment_metadata()),
     prettyversion(vcstools.GitClient.get_environment_metadata()),
     prettyversion(vcstools.TarClient.get_environment_metadata()),
     prettyversion(vcstools.BzrClient.get_environment_metadata()))


def cmd_status(config, localnames=None, untracked=False):
    """
    calls SCM status for all SCM entries in config, relative to path

    :returns: List of dict {element: ConfigElement, diff: diffstring}
    :param untracked: also show files not added to the SCM
    :raises MultiProjectException: on plenty of errors
    """
    class StatusRetriever():

        def __init__(self, element, path, untracked):
            self.element = element
            self.path = path
            self.untracked = untracked

        def do_work(self):
            path_spec = self.element.get_path_spec()
            scmtype = path_spec.get_scmtype()
            status = self.element.get_status(self.path, self.untracked)
            # align other scm output to svn
            columns = -1
            if scmtype == "git":
                columns = 3
            elif scmtype == "hg":
                columns = 2
            elif scmtype == "bzr":
                columns = 4
            if columns > -1 and status is not None:
                status_aligned = ''
                for line in status.splitlines():
                    status_aligned = "%s%s%s\n"%(status_aligned,
                                                 line[:columns].ljust(8),
                                                 line[columns:])
                status = status_aligned
            return {'status': status}

    path = config.get_base_path()
    # call SCM info in separate threads
    elements = config.get_config_elements()
    work = DistributedWork(len(elements))
    elements = select_elements(config, localnames)
    for element in elements:
        if element.is_vcs_element():
            work.add_thread(StatusRetriever(element, path, untracked))
    outputs = work.run()
    return outputs


def cmd_diff(config, localnames=None):
    """
    calls SCM diff for all SCM entries in config, relative to path

    :returns: List of dict {element: ConfigElement, diff: diffstring}
    :raises MultiProjectException: on plenty of errors
    """
    class DiffRetriever():

        def __init__(self, element, path):
            self.element = element
            self.path = path

        def do_work(self):
            return {'diff': self.element.get_diff(self.path)}

    path = config.get_base_path()
    elements = config.get_config_elements()
    work = DistributedWork(len(elements))
    elements = select_elements(config, localnames)
    for element in elements:
        if element.is_vcs_element():
            work.add_thread(DiffRetriever(element, path))
    outputs = work.run()
    return outputs


def cmd_install_or_update(
    config,
    backup_path=None,
    mode='abort',
    robust=False,
    localnames=None,
    num_threads=1,
    verbose=False):
    """
    performs many things, generally attempting to make
    the local filesystem look like what the config specifies,
    pulling from remote sources the most recent changes.

    The command may have stdin user interaction (TODO abstract)

    :param backup_path: if and where to backup trees before deleting them
    :param robust: proceed to next element even when one element fails
    :returns: True on Success
    :raises MultiProjectException: on plenty of errors
    """
    success = True
    if not os.path.exists(config.get_base_path()):
        os.mkdir(config.get_base_path())
    # Prepare install operation check filesystem and ask user
    preparation_reports = []
    elements = select_elements(config, localnames)
    for t in elements:
        abs_backup_path = None
        if backup_path is not None:
            abs_backup_path = os.path.join(config.get_base_path(), backup_path)
        try:
            preparation_report = t.prepare_install(backup_path=abs_backup_path,
                                                   arg_mode=mode,
                                                   robust=robust)
            if preparation_report is not None:
                if preparation_report.abort:
                    raise MultiProjectException("Aborting install because of %s"%preparation_report.error)
                if not preparation_report.skip:
                    preparation_reports.append(preparation_report)
                else:
                    if preparation_report.error is not None:
                        print("Skipping install of %s because: %s"%(preparation_report.config_element.get_local_name(),
                                                                    preparation_report.error))
        except MultiProjectException as ex:
            fail_str = "Failed to install tree '%s'\n %s"%(t.get_path(), ex)
            if robust:
                success = False
                print("Continuing despite %s"%fail_str)
            else:
                raise MultiProjectException(fail_str)

    class Installer():

        def __init__(self, report):
            self.element = report.config_element
            self.report = report

        def do_work(self):
            self.element.install(checkout=self.report.checkout,
                                 backup=self.report.backup,
                                 backup_path=self.report.backup_path,
                                 verbose=self.report.verbose)
            return {}

    work = DistributedWork(len(preparation_reports), num_threads, silent=False)
    for report in preparation_reports:
        report.verbose = verbose
        thread = Installer(report)
        work.add_thread(thread)

    try:
        work.run()
    except MultiProjectException as e:
        print ("Exception caught during install: %s"%e)
        success = False
        if not robust:
            raise e
    return success
    # TODO go back and make sure that everything in options.path is described
    # in the yaml, and offer to delete otherwise? not sure, but it could go here


def cmd_snapshot(config, localnames=None):
    elements = select_elements(config, localnames)
    source_aggregate = []
    for element in elements:
        source = element.get_versioned_path_spec().get_legacy_yaml()
        source_aggregate.append(source)
    return source_aggregate


def cmd_info(config, localnames=None):
    """This function compares what should be (config_file) with what is
    (directories) and returns a list of dictionary giving each local
    path and all the state information about it available.
    """

    class InfoRetriever():

        def __init__(self, element, path):
            self.element = element
            self.path = path

        def do_work(self):
            localname = ""
            scm = None
            uri = ""
            curr_uri = None
            exists = False
            version = ""# what is given in config file
            modified = ""
            actualversion = "" # revision number of version
            specversion = "" # actual revision number
            localname = self.element.get_local_name()
            path = self.element.get_path() or localname

            if localname is None or localname == "":
                raise MultiProjectException("Missing local-name in element: %s"%self.element)
            abs_path = normabspath(path, self.path)
            if (os.path.exists(abs_path)):
                exists = True
            if self.element.is_vcs_element():
                if not exists:
                    path_spec = self.element.get_path_spec()
                    version = path_spec.get_version()
                else:
                    path_spec = self.element.get_versioned_path_spec()
                    version = path_spec.get_version()
                    curr_uri = path_spec.get_curr_uri()
                    status = self.element.get_status(self.path)
                    if (status is not None and
                        status.strip() != ''):
                        modified = True
                    specversion = path_spec.get_revision()
                    if (version is not None and
                        version.strip() != '' and
                        (specversion is None or specversion.strip() == '')):

                        specversion = '"%s"'%version
                    actualversion = path_spec.get_current_revision()
                scm = path_spec.get_scmtype()
                uri = path_spec.get_uri()
            return {'scm': scm,
                    'exists': exists,
                    'localname': localname,
                    'path': path,
                    'uri': uri,
                    'curr_uri': curr_uri,
                    'version': version,
                    'specversion': specversion,
                    'actualversion': actualversion,
                    'modified': modified,
                    'properties': self.element.get_properties()}
    path = config.get_base_path()
    # call SCM info in separate threads
    elements = config.get_config_elements()
    work = DistributedWork(len(elements))
    elements = select_elements(config, localnames)
    for element in elements:
        if element.get_properties() is None or not 'setup-file' in element.get_properties():
            work.add_thread(InfoRetriever(element, path))
    outputs = work.run()
    return outputs



from setuptools import setup

import imp


def get_version():
    ver_file = None
    try:
        ver_file, pathname, description = imp.find_module('__version__', ['src/rosinstall'])
        vermod = imp.load_module('__version__', ver_file, pathname, description)
        version = vermod.version
        return version
    finally:
        if ver_file is not None:
            ver_file.close()


setup(name='rosinstall',
      version=get_version(),
      packages=['rosinstall'],
      package_dir={'': 'src'},
      install_requires=['vcstools', 'pyyaml'],
      scripts=["scripts/rosinstall", "scripts/roslocate", "scripts/rosws", "scripts/rosco"],
      author="Tully Foote",
      author_email="tfoote@willowgarage.com",
      url="http://www.ros.org/wiki/rosinstall",
      download_url="http://pr.willowgarage.com/downloads/rosinstall/",
      keywords=["ROS"],
      classifiers=["Programming Language :: Python",
                   "License :: OSI Approved :: BSD License"],
      description="The installer for ROS",
      long_description="""\
The installer for ROS
""",
      license="BSD")
